@@ -98,10 +98,25 @@ static int umount_all(struct autofs_point *ap, int force);
 
 extern struct master *master_list;
 
+static int is_remote_fstype(unsigned int fs_type)
+{
+	int ret = 0;
+	switch (fs_type) {
+	case SMB_SUPER_MAGIC:
+	case CIFS_MAGIC_NUMBER:
+	case NCP_SUPER_MAGIC:
+	case NFS_SUPER_MAGIC:
+		ret = 1;
+		break;
+	};
+	return ret;
+}
+
 static int do_mkdir(const char *parent, const char *path, mode_t mode)
 {
 	int status;
-	struct stat st;
+	mode_t mask;
+	struct stat st, root;
 	struct statfs fs;
 
 	/* If path exists we're done */
@@ -114,24 +129,37 @@ static int do_mkdir(const char *parent, const char *path, mode_t mode)
 	}
 
 	/*
-	 * If we're trying to create a directory within an autofs fs
-	 * or the path is contained in a localy mounted fs go ahead.
+	 * We don't want to create the path on a remote file system
+	 * unless it's the root file system.
+	 * An empty parent means it's the root directory and always ok.
 	 */
-	status = -1;
-	if (*parent)
+	if (*parent) {
 		status = statfs(parent, &fs);
-	if ((status != -1 && fs.f_type == (__SWORD_TYPE) AUTOFS_SUPER_MAGIC) ||
-	    contained_in_local_fs(path)) {
-		mode_t mask = umask(0022);
-		int ret = mkdir(path, mode);
-		(void) umask(mask);
-		if (ret == -1) {
-			errno = EACCES;
-			return 0;
+		if (status == -1)
+			goto fail;
+
+		if (is_remote_fstype(fs.f_type)) {
+			status = stat(parent, &st);
+			if (status == -1)
+				goto fail;
+
+			status = stat("/", &root);
+			if (status == -1)
+				goto fail;
+
+			if (st.st_dev != root.st_dev)
+				goto fail;
 		}
-		return 1;
 	}
 
+	mask = umask(0022);
+	status = mkdir(path, mode);
+	(void) umask(mask);
+	if (status == -1)
+		goto fail;
+
+	return 1;
+fail:
 	errno = EACCES;
 	return 0;
 }
@@ -462,7 +490,7 @@ static int umount_subtree_mounts(struct autofs_point *ap, const char *path, unsi
 {
 	struct mapent_cache *mc;
 	struct mapent *me;
-	unsigned int is_mm_root;
+	unsigned int is_mm_root = 0;
 	int left;
 
 	me = lookup_source_mapent(ap, path, LKP_DISTINCT);
@@ -474,22 +502,22 @@ static int umount_subtree_mounts(struct autofs_point *ap, const char *path, unsi
 			ind_key++;
 
 		me = lookup_source_mapent(ap, ind_key, LKP_NORMAL);
-		if (!me)
-			return 0;
 	}
 
-	mc = me->mc;
-	is_mm_root = (me->multi == me);
+	if (me) {
+		mc = me->mc;
+		is_mm_root = (me->multi == me);
+	}
 
 	left = 0;
 
-	pthread_cleanup_push(cache_lock_cleanup, mc);
-
-	if (me->multi) {
+	if (me && me->multi) {
 		char root[PATH_MAX];
 		char *base;
 		int cur_state;
 
+		pthread_cleanup_push(cache_lock_cleanup, mc);
+
 		if (!strchr(me->multi->key, '/'))
 			/* Indirect multi-mount root */
 			/* sprintf okay - if it's mounted, it's
@@ -512,10 +540,15 @@ static int umount_subtree_mounts(struct autofs_point *ap, const char *path, unsi
 			left++;
 		}
 		cache_multi_unlock(me->parent);
+		if (ap->entry->maps &&
+		    (ap->entry->maps->flags & MAP_FLAG_FORMAT_AMD))
+			cache_pop_mapent(me);
 		pthread_setcancelstate(cur_state, NULL);
+		pthread_cleanup_pop(0);
 	}
 
-	pthread_cleanup_pop(1);
+	if (me)
+		cache_unlock(mc);
 
 	if (left || is_autofs_fs)
 		return left;
@@ -525,14 +558,37 @@ static int umount_subtree_mounts(struct autofs_point *ap, const char *path, unsi
 	 * it already to ensure it's ok to remove any offset triggers.
 	 */
 	if (!is_mm_root && is_mounted(_PATH_MOUNTED, path, MNTS_REAL)) {
+		struct amd_entry *entry;
 		debug(ap->logopt, "unmounting dir = %s", path);
 		if (umount_ent(ap, path) &&
 		    is_mounted(_PATH_MOUNTED, path, MNTS_REAL)) {
 			warn(ap->logopt, "could not umount dir %s", path);
 			left++;
+			goto done;
 		}
-	}
 
+		/* Check for an external mount and umount if possible */
+		mounts_mutex_lock(ap);
+		entry = __master_find_amdmount(ap, path);
+		if (!entry) {
+			mounts_mutex_unlock(ap);
+			goto done;
+		}
+		list_del(&entry->entries);
+		mounts_mutex_unlock(ap);
+		if (ext_mount_remove(&entry->ext_mount, entry->fs)) {
+			if (umount_ent(ap, entry->fs))
+				debug(ap->logopt,
+				      "failed to umount external mount %s",
+				      entry->fs);
+			else
+				debug(ap->logopt,
+				      "umounted external mount %s",
+				      entry->fs);
+		}
+		free_amd_entry(entry);
+	}
+done:
 	return left;
 }
 
@@ -541,10 +597,60 @@ static int umount_subtree_mounts(struct autofs_point *ap, const char *path, unsi
 int umount_multi(struct autofs_point *ap, const char *path, int incl)
 {
 	int is_autofs_fs;
+	struct stat st;
 	int left;
 
 	debug(ap->logopt, "path %s incl %d", path, incl);
 
+	/* If path is a mount it can't be a symlink */
+	if (is_mounted(_PATH_MOUNTED, path, MNTS_ALL))
+		goto real_mount;
+
+	if (lstat(path, &st)) {
+		warn(ap->logopt,
+		     "failed to stat directory or symlink %s", path);
+		return 1;
+	}
+
+	/* if this is a symlink we can handle it now */
+	if (S_ISLNK(st.st_mode)) {
+		struct amd_entry *entry;
+		if (st.st_dev != ap->dev) {
+			crit(ap->logopt,
+			     "symlink %s has the wrong device, "
+			     "possible race condition", path);
+			return 1;
+		}
+		debug(ap->logopt, "removing symlink %s", path);
+		if (unlink(path)) {
+			error(ap->logopt,
+			      "failed to remove symlink %s", path);
+			return 1;
+		}
+		/* Check for an external mount and attempt umount if needed */
+		mounts_mutex_lock(ap);
+		entry = __master_find_amdmount(ap, path);
+		if (!entry) {
+			mounts_mutex_unlock(ap);
+			return 0;
+		}
+		list_del(&entry->entries);
+		mounts_mutex_unlock(ap);
+		if (ext_mount_remove(&entry->ext_mount, entry->fs)) {
+			if (umount_ent(ap, entry->fs))
+				debug(ap->logopt,
+				      "failed to umount external mount %s",
+				      entry->fs);
+			else
+				debug(ap->logopt,
+				      "umounted external mount %s",
+				      entry->fs);
+		}
+		free_amd_entry(entry);
+		return 0;
+	}
+
+real_mount:
 	is_autofs_fs = 0;
 	if (master_find_submount(ap, path))
 		is_autofs_fs = 1;
@@ -556,7 +662,7 @@ int umount_multi(struct autofs_point *ap, const char *path, int incl)
 	 * parent may have mounted over top of us.
 	 */
 	if (ap->submount)
-		left += umount_subtree_mounts(ap->parent, path, is_autofs_fs);
+		left += umount_subtree_mounts(ap->parent, path, 1);
 
 	left += umount_subtree_mounts(ap, path, is_autofs_fs);
 
@@ -890,6 +996,8 @@ static int get_pkt(struct autofs_point *ap, union autofs_v5_packet_union *pkt)
 	struct pollfd fds[3];
 	int pollfds = 3;
 	char buf[MAX_ERR_BUF];
+	size_t read;
+	char *estr;
 
 	fds[0].fd = ap->pipefd;
 	fds[0].events = POLLIN;
@@ -902,7 +1010,6 @@ static int get_pkt(struct autofs_point *ap, union autofs_v5_packet_union *pkt)
 
 	for (;;) {
 		if (poll(fds, pollfds, -1) == -1) {
-			char *estr;
 			if (errno == EINTR)
 				continue;
 			estr = strerror_r(errno, buf, MAX_ERR_BUF);
@@ -921,7 +1028,13 @@ static int get_pkt(struct autofs_point *ap, union autofs_v5_packet_union *pkt)
 
 			state_pipe = ap->state_pipe[0];
 
-			if (fullread(state_pipe, &next_state, read_size)) {
+			read = fullread(state_pipe, &next_state, read_size);
+			if (read) {
+				estr = strerror_r(errno, buf, MAX_ERR_BUF);
+				error(ap->logopt,
+				      "read error on state pipe, "
+				      "read %u, error %s",
+				      read, estr);
 				st_mutex_unlock();
 				continue;
 			}
@@ -932,8 +1045,17 @@ static int get_pkt(struct autofs_point *ap, union autofs_v5_packet_union *pkt)
 				return -1;
 		}
 
-		if (fds[0].revents & POLLIN)
-			return fullread(ap->pipefd, pkt, kpkt_len);
+		if (fds[0].revents & POLLIN) {
+			read = fullread(ap->pipefd, pkt, kpkt_len);
+			if (read) {
+				estr = strerror_r(errno, buf, MAX_ERR_BUF);
+				error(ap->logopt,
+				      "read error on request pipe, "
+				      "read %u, expected %u error %s",
+				       read, kpkt_len, estr);
+			}
+			return read;
+		}
 
 		if (fds[2].fd != -1 && fds[2].revents & POLLIN) {
 			debug(ap->logopt, "message pending on control fifo.");
@@ -1285,10 +1407,9 @@ static void *do_read_master(void *arg)
 	return NULL;
 }
 
-static int do_hup_signal(struct master *master)
+static int do_hup_signal(struct master *master, time_t age)
 {
 	unsigned int logopt = master->logopt;
-	time_t age = time(NULL);
 	pthread_t thid;
 	int status;
 
@@ -1377,7 +1498,7 @@ static void *statemachine(void *arg)
 			break;
 
 		case SIGHUP:
-			do_hup_signal(master_list);
+			do_hup_signal(master_list, monotonic_time(NULL));
 			break;
 
 		default:
@@ -1477,9 +1598,15 @@ static void handle_mounts_cleanup(void *arg)
 		clean = 1;
 
 	if (submount) {
+		struct amd_entry *am;
 		/* We are finishing up */
 		ap->parent->submnt_count--;
 		list_del_init(&ap->mounts);
+		am = __master_find_amdmount(ap->parent, ap->path);
+		if (am) {
+			list_del_init(&am->entries);
+			free_amd_entry(am);
+		}
 	}
 
 	/* Don't signal the handler if we have already done so */
@@ -1560,6 +1687,99 @@ static void submount_source_unlock_nested(struct autofs_point *ap)
 	master_source_unlock(parent->entry);
 }
 
+int handle_mounts_exit(struct autofs_point *ap)
+{
+	int ret, cur_state;
+
+	/*
+	 * If we're a submount we need to ensure our parent
+	 * doesn't try to mount us again until our shutdown
+	 * is complete and that any outstanding mounts are
+	 * completed before we try to shutdown.
+	 */
+	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cur_state);
+
+	master_mutex_lock();
+
+	if (!ap->submount)
+		master_source_writelock(ap->entry);
+	else {
+		/*
+		 * If a mount request arrives before the locks are
+		 * aquired just return to ready state.
+		 */
+		ret = submount_source_writelock_nested(ap);
+		if (ret) {
+			warn(ap->logopt,
+			     "can't shutdown submount: mount in progress");
+			/* Return to ST_READY is done immediately */
+			st_add_task(ap, ST_READY);
+			master_mutex_unlock();
+			pthread_setcancelstate(cur_state, NULL);
+			return 0;
+		}
+	}
+
+	if (ap->state != ST_SHUTDOWN) {
+		if (!ap->submount)
+			alarm_add(ap, ap->exp_runfreq);
+		/* Return to ST_READY is done immediately */
+		st_add_task(ap, ST_READY);
+		if (ap->submount)
+			submount_source_unlock_nested(ap);
+		else
+			master_source_unlock(ap->entry);
+		master_mutex_unlock();
+
+		pthread_setcancelstate(cur_state, NULL);
+		return 0;
+	}
+
+	alarm_delete(ap);
+	st_remove_tasks(ap);
+	st_wait_task(ap, ST_ANY, 0);
+
+	/*
+	 * For a direct mount map all mounts have already gone
+	 * by the time we get here and since we only ever
+	 * umount direct mounts at shutdown there is no need
+	 * to check for possible recovery.
+	 */
+	if (ap->type == LKP_DIRECT) {
+		umount_autofs(ap, NULL, 1);
+		handle_mounts_cleanup(ap);
+		return 1;
+	}
+
+	/*
+	 * If umount_autofs returns non-zero it wasn't able
+	 * to complete the umount and has left the mount intact
+	 * so we can continue. This can happen if a lookup
+	 * occurs while we're trying to umount.
+	 */
+	ret = umount_autofs(ap, NULL, 1);
+	if (!ret) {
+		handle_mounts_cleanup(ap);
+		return 1;
+	}
+
+	/* Failed shutdown returns to ready */
+	warn(ap->logopt, "can't shutdown: filesystem %s still busy", ap->path);
+	if (!ap->submount)
+		alarm_add(ap, ap->exp_runfreq);
+	/* Return to ST_READY is done immediately */
+	st_add_task(ap, ST_READY);
+	if (ap->submount)
+		submount_source_unlock_nested(ap);
+	else
+		master_source_unlock(ap->entry);
+	master_mutex_unlock();
+
+	pthread_setcancelstate(cur_state, NULL);
+
+	return 0;
+}
+
 void *handle_mounts(void *arg)
 {
 	struct startup_cond *suc;
@@ -1615,99 +1835,24 @@ void *handle_mounts(void *arg)
 
 	pthread_setcancelstate(cancel_state, NULL);
 
-	while (ap->state != ST_SHUTDOWN) {
+	while (1) {
 		if (handle_packet(ap)) {
-			int ret, cur_state;
-
-			/*
-			 * If we're a submount we need to ensure our parent
-			 * doesn't try to mount us again until our shutdown
-			 * is complete and that any outstanding mounts are
-			 * completed before we try to shutdown.
-			 */
-			pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cur_state);
-
-			master_mutex_lock();
-
-			if (ap->submount) {
-				/*
-				 * If a mount request arrives before the locks are
-				 * aquired just return to ready state.
-				 */
-				ret = submount_source_writelock_nested(ap);
-				if (ret) {
-					warn(ap->logopt,
-					     "can't shutdown submount: mount in progress");
-					/* Return to ST_READY is done immediately */
-					st_add_task(ap, ST_READY);
-					master_mutex_unlock();
-					pthread_setcancelstate(cur_state, NULL);
-					continue;
-				}
-			} else
-				master_source_writelock(ap->entry);
-
-			if (ap->state != ST_SHUTDOWN) {
-				if (!ap->submount)
-					alarm_add(ap, ap->exp_runfreq);
-				/* Return to ST_READY is done immediately */
-				st_add_task(ap, ST_READY);
-				if (ap->submount)
-					submount_source_unlock_nested(ap);
-				else
-					master_source_unlock(ap->entry);
-				master_mutex_unlock();
-
-				pthread_setcancelstate(cur_state, NULL);
-				continue;
-			}
-
-			alarm_delete(ap);
-			st_remove_tasks(ap);
-			st_wait_task(ap, ST_ANY, 0);
-
-			/*
-			 * For a direct mount map all mounts have already gone
-			 * by the time we get here and since we only ever
-			 * umount direct mounts at shutdown there is no need
-			 * to check for possible recovery.
-			 */
-			if (ap->type == LKP_DIRECT) {
-				umount_autofs(ap, NULL, 1);
+			if (handle_mounts_exit(ap))
 				break;
-			}
+		}
 
-			/*
-			 * If umount_autofs returns non-zero it wasn't able
-			 * to complete the umount and has left the mount intact
-			 * so we can continue. This can happen if a lookup
-			 * occurs while we're trying to umount.
-			 */
-			ret = umount_autofs(ap, NULL, 1);
-			if (!ret)
+		/* If we get here a packet has been received and handled
+		 * and the autofs mount point has not been shutdown. But
+		 * if the autofs mount point has been set to ST_SHUTDOWN
+		 * we should attempt to perform the shutdown cleanup and
+		 * exit if successful.
+		 */
+		if (ap->state == ST_SHUTDOWN) {
+			if (handle_mounts_exit(ap))
 				break;
-
-			/* Failed shutdown returns to ready */
-			warn(ap->logopt,
-			     "can't shutdown: filesystem %s still busy",
-			     ap->path);
-			if (!ap->submount)
-				alarm_add(ap, ap->exp_runfreq);
-			/* Return to ST_READY is done immediately */
-			st_add_task(ap, ST_READY);
-			if (ap->submount)
-				submount_source_unlock_nested(ap);
-			else
-				master_source_unlock(ap->entry);
-			master_mutex_unlock();
-
-			pthread_setcancelstate(cur_state, NULL);
-
 		}
 	}
 
-	handle_mounts_cleanup(ap);
-
 	return NULL;
 }
 
@@ -1750,9 +1895,6 @@ static void usage(void)
 		"	-C --dont-check-daemon\n"
 		"			don't check if daemon is already running\n"
 		"	-F --force	forceably clean up known automounts at start\n"
-		"	-M --master-wait n\n"
-		"			maximum time to wait for the master map to\n"
-		"			become available at start up\n"
 		"	-V --version	print version, build config and exit\n"
 		, program);
 }
@@ -1917,61 +2059,16 @@ static void remove_empty_args(char **argv, int *argc)
 	*argc = j;
 }
 
-static int do_master_read_master(struct master *master, int wait)
-{
-	sigset_t signalset;
-	/* Wait must be at least 1 second */
-	unsigned int retry_wait = 2;
-	unsigned int elapsed = 0;
-	int max_wait = wait;
-	int ret = 0;
-	time_t age;
-
-	sigemptyset(&signalset);
-	sigaddset(&signalset, SIGTERM);
-	sigaddset(&signalset, SIGINT);
-	sigaddset(&signalset, SIGHUP);
-	sigprocmask(SIG_UNBLOCK, &signalset, NULL);
-
-	while (1) {
-		struct timespec t = { retry_wait, 0 };
-
-		age = time(NULL);
-		if (master_read_master(master, age, 0)) {
-			ret = 1;
-			break;
-		}
-
-		if (nanosleep(&t, NULL) == -1)
-			break;
-
-		if (max_wait > 0) {
-			elapsed += retry_wait;
-			if (elapsed >= max_wait) {
-				logmsg("problem reading master map, "
-					"maximum wait exceeded");
-				break;
-			}
-		}
-	}
-
-	sigprocmask(SIG_BLOCK, &signalset, NULL);
-
-	return ret;
-}
-
 int main(int argc, char *argv[])
 {
 	int res, opt, status;
 	int logpri = -1;
 	unsigned ghost, logging, daemon_check;
 	unsigned dumpmaps, foreground, have_global_options;
-	unsigned master_read;
-	int master_wait;
 	time_t timeout;
-	time_t age = time(NULL);
+	time_t age = monotonic_time(NULL);
 	struct rlimit rlim;
-	const char *options = "+hp:t:vmdD:fVrO:l:n:CFM";
+	const char *options = "+hp:t:vmdD:fVrO:l:n:CF";
 	static const struct option long_options[] = {
 		{"help", 0, 0, 'h'},
 		{"pid-file", 1, 0, 'p'},
@@ -1988,7 +2085,6 @@ int main(int argc, char *argv[])
 		{"set-log-priority", 1, 0, 'l'},
 		{"dont-check-daemon", 0, 0, 'C'},
 		{"force", 0, 0, 'F'},
-		{"master-wait", 1, 0, 'M'},
 		{0, 0, 0, 0}
 	};
 
@@ -2009,7 +2105,6 @@ int main(int argc, char *argv[])
 	nfs_mount_uses_string_options = check_nfs_mount_version(&vers, &check);
 
 	kpkt_len = get_kpkt_len();
-	master_wait = defaults_get_master_wait();
 	timeout = defaults_get_timeout();
 	ghost = defaults_get_browse_mode();
 	logging = defaults_get_logging();
@@ -2069,10 +2164,6 @@ int main(int argc, char *argv[])
 			dumpmaps = 1;
 			break;
 
-		case 'M':
-			master_wait = getnumopt(optarg, opt);
-			break;
-
 		case 'O':
 			if (!have_global_options) {
 				global_options = strdup(optarg);
@@ -2364,18 +2455,18 @@ int main(int argc, char *argv[])
 	void *dh_tirpc = dlopen("libtirpc.so", RTLD_NOW);
 	if (!dh_tirpc)
 		dh_tirpc = dlopen("libtirpc.so.1", RTLD_NOW);
+	if (!dh_tirpc)
+		dh_tirpc = dlopen("libtirpc.so.3", RTLD_NOW);
 #endif
 
-	master_read = master_read_master(master_list, age, 0);
-	if (!master_read) {
-		if (foreground)
-			logerr("%s: failed to read master map, "
-			       "will retry!",
-			       program);
-		else
-			logerr("%s: failed to read master map, "
-			       "will retry in background!",
-			       program);
+	if (!master_read_master(master_list, age, 0)) {
+		master_kill(master_list);
+		*pst_stat = 3;
+		res = write(start_pipefd[1], pst_stat, sizeof(*pst_stat));
+		close(start_pipefd[1]);
+		release_flag_file();
+		macro_free_global_table();
+		exit(3);
 	}
 
 	/*
@@ -2388,20 +2479,6 @@ int main(int argc, char *argv[])
 	res = write(start_pipefd[1], pst_stat, sizeof(*pst_stat));
 	close(start_pipefd[1]);
 
-	if (!master_read) {
-		/*
-		 * Read master map, waiting until it is available, unless
-		 * a signal is received, in which case exit returning an
-		 * error.
-		 */
-		if (!do_master_read_master(master_list, -1)) {
-			logerr("%s: failed to read master map!", program);
-			master_kill(master_list);
-			release_flag_file();
-			exit(3);
-		}
-	}
-
 	state_mach_thid = pthread_self();
 	statemachine(NULL);
 
@@ -2411,6 +2488,7 @@ int main(int argc, char *argv[])
 		unlink(pid_file);
 		pid_file = NULL;
 	}
+	defaults_conf_release();
 	closelog();
 	release_flag_file();
 	macro_free_global_table();
