@@ -25,51 +25,6 @@
 #include "automount.h"
 #include "nsswitch.h"
 
-static int check_nss_result(struct nss_source *this, enum nsswitch_status result)
-{
-	enum nsswitch_status status;
-	struct nss_action a;
-
-	/* Check if we have negated actions */
-	for (status = 0; status < NSS_STATUS_MAX; status++) {
-		a = this->action[status];
-		if (a.action == NSS_ACTION_UNKNOWN)
-			continue;
-
-		if (a.negated && result != status) {
-			if (a.action == NSS_ACTION_RETURN) {
-				if (result == NSS_STATUS_SUCCESS)
-					return 1;
-				else
-					return 0;
-			}
-		}
-	}
-
-	a = this->action[result];
-
-	/* Check if we have other actions for this status */
-	switch (result) {
-	case NSS_STATUS_SUCCESS:
-		if (a.action == NSS_ACTION_CONTINUE)
-			break;
-		return 1;
-
-	case NSS_STATUS_NOTFOUND:
-	case NSS_STATUS_UNAVAIL:
-	case NSS_STATUS_TRYAGAIN:
-		if (a.action == NSS_ACTION_RETURN) {
-			return 0;
-		}
-		break;
-
-	default:
-		break;
-	}
-
-	return -1;
-}
-
 static void nsslist_cleanup(void *arg)
 {
 	struct list_head *nsslist = (struct list_head *) arg;
@@ -89,9 +44,9 @@ static int do_read_master(struct master *master, char *type, time_t age)
 	argv[0] = master->name;
 	argv[1] = NULL;
 
-	lookup = open_lookup(type, "", NULL, argc, argv);
-	if (!lookup)
-		return NSS_STATUS_UNAVAIL;
+	status = open_lookup(type, "", NULL, argc, argv, &lookup);
+	if (status != NSS_STATUS_SUCCESS)
+		return status;
 
 	status = lookup->lookup_read_master(master, age, lookup->context);
 
@@ -233,8 +188,8 @@ int lookup_nss_read_master(struct master *master, time_t age)
 			memset(source, 0, 10);
 			if ((!strncmp(name, "file", 4) &&
 				 (name[4] == ',' || name[4] == ':')) ||
-			    (!strncmp(name, "yp", 3) &&
-				 (name[3] == ',' || name[3] == ':')) ||
+			    (!strncmp(name, "yp", 2) &&
+				 (name[2] == ',' || name[2] == ':')) ||
 			    (!strncmp(name, "nis", 3) &&
 				 (name[3] == ',' || name[3] == ':')) ||
 			    (!strncmp(name, "nisplus", 7) &&
@@ -345,16 +300,27 @@ static int do_read_map(struct autofs_point *ap, struct map_source *map, time_t a
 	struct lookup_mod *lookup;
 	int status;
 
-	lookup = open_lookup(map->type, "", map->format, map->argc, map->argv);
-	if (!lookup) {
-		debug(ap->logopt, "lookup module %s failed", map->type);
-		return NSS_STATUS_UNAVAIL;
-	}
-
+	lookup = NULL;
 	master_source_writelock(ap->entry);
-	if (map->lookup)
-		close_lookup(map->lookup);
-	map->lookup = lookup;
+	if (!map->lookup) {
+		status = open_lookup(map->type, "", map->format,
+				     map->argc, map->argv, &lookup);
+		if (status != NSS_STATUS_SUCCESS) {
+			master_source_unlock(ap->entry);
+			debug(ap->logopt,
+			      "lookup module %s open failed", map->type);
+			return status;
+		}
+		map->lookup = lookup;
+	} else {
+		lookup = map->lookup;
+		status = lookup->lookup_reinit(map->format,
+					       map->argc, map->argv,
+					       &lookup->context);
+		if (status)
+			warn(ap->logopt,
+			     "lookup module %s reinit failed", map->type);
+	}
 	master_source_unlock(ap->entry);
 
 	if (!map->stale)
@@ -782,12 +748,12 @@ int do_lookup_mount(struct autofs_point *ap, struct map_source *map, const char
 	int status;
 
 	if (!map->lookup) {
-		lookup = open_lookup(map->type, "",
-				     map->format, map->argc, map->argv);
-		if (!lookup) {
+		status = open_lookup(map->type, "",
+				     map->format, map->argc, map->argv, &lookup);
+		if (status != NSS_STATUS_SUCCESS) {
 			debug(ap->logopt,
-			      "lookup module %s failed", map->type);
-			return NSS_STATUS_UNAVAIL;
+			      "lookup module %s open failed", map->type);
+			return status;
 		}
 		map->lookup = lookup;
 	}
@@ -843,7 +809,7 @@ static int lookup_amd_instance(struct autofs_point *ap,
 		return NSS_STATUS_UNKNOWN;
 	}
 
-	if (entry->opts) {
+	if (entry->opts && *entry->opts) {
 		argv[0] = entry->opts;
 		argv[1] = NULL;
 		pargv = argv;
@@ -875,7 +841,7 @@ static int lookup_name_file_source_instance(struct autofs_point *ap, struct map_
 	struct map_source *instance;
 	char src_file[] = "file";
 	char src_prog[] = "program";
-	time_t age = time(NULL);
+	time_t age = monotonic_time(NULL);
 	struct stat st;
 	char *type, *format;
 
@@ -915,7 +881,7 @@ static int lookup_name_source_instance(struct autofs_point *ap, struct map_sourc
 {
 	struct map_source *instance;
 	const char *format;
-	time_t age = time(NULL);
+	time_t age = monotonic_time(NULL);
 
 	if (*name == '/' && map->flags & MAP_FLAG_FORMAT_AMD)
 		return lookup_amd_instance(ap, map, name, name_len);
@@ -1057,6 +1023,10 @@ static void update_negative_cache(struct autofs_point *ap, struct map_source *so
 	if (source && source->depth)
 		return;
 
+	/* Don't update the wildcard */
+	if (strlen(name) == 1 && *name == '*')
+		return;
+
 	/* Have we recorded the lookup fail for negative caching? */
 	me = lookup_source_mapent(ap, name, LKP_DISTINCT);
 	if (me)
@@ -1066,8 +1036,15 @@ static void update_negative_cache(struct autofs_point *ap, struct map_source *so
 		 */
 		cache_unlock(me->mc);
 	else {
-		/* Notify only once after fail */
-		logmsg("key \"%s\" not found in map source(s).", name);
+		if (!defaults_disable_not_found_message()) {
+			/* This really should be a warning but the original
+			 * request for this needed it to be unconditional.
+			 * That produces, IMHO, unnecessary noise in the log
+			 * so a configuration option has been added to provide
+			 * the ability to turn it off.
+			 */
+			logmsg("key \"%s\" not found in map source(s).", name);
+		}
 
 		/* Doesn't exist in any source, just add it somewhere */
 		if (source)
@@ -1075,7 +1052,7 @@ static void update_negative_cache(struct autofs_point *ap, struct map_source *so
 		else
 			map = entry->maps;
 		if (map) {
-			time_t now = time(NULL);
+			time_t now = monotonic_time(NULL);
 			int rv = CHE_FAIL;
 
 			cache_writelock(map->mc);
@@ -1103,7 +1080,7 @@ int lookup_nss_mount(struct autofs_point *ap, struct map_source *source, const c
 	struct nss_source *this;
 	struct map_source *map;
 	enum nsswitch_status status;
-	int result = 0;
+	int result = NSS_STATUS_UNKNOWN;
 
 	/*
 	 * For each map source (ie. each entry for the mount
@@ -1122,6 +1099,7 @@ int lookup_nss_mount(struct autofs_point *ap, struct map_source *source, const c
 		 * the map entry was last updated.
 		 */
 		if (entry->age > map->age) {
+			status = NSS_STATUS_UNAVAIL;
 			map = map->next;
 			continue;
 		}
@@ -1143,6 +1121,7 @@ int lookup_nss_mount(struct autofs_point *ap, struct map_source *source, const c
 				char *tmp = strdup("ldap");
 				if (!tmp) {
 					map = map->next;
+					status = NSS_STATUS_TRYAGAIN;
 					continue;
 				}
 				map->type = tmp;
@@ -1175,6 +1154,7 @@ int lookup_nss_mount(struct autofs_point *ap, struct map_source *source, const c
 			    !strcmp(this->source, "sss")) {
 				warn(ap->logopt,
 				     "source sss is not available for amd maps.");
+				result = NSS_STATUS_UNAVAIL;
 				continue;
 			}
 
