@@ -144,7 +144,7 @@ void expire_cleanup(void *arg)
 					ap->submount = 2;
 			}
 
-			if (!ap->submount)
+			if (ap->state == ST_EXPIRE && !ap->submount)
 				alarm_add(ap, ap->exp_runfreq);
 
 			/* FALLTHROUGH */
@@ -330,13 +330,9 @@ static void do_readmap_cleanup(void *arg)
 	ap = ra->ap;
 
 	st_mutex_lock();
-
 	ap->readmap_thread = 0;
 	st_set_done(ap);
-	if (!ap->submount)
-		alarm_add(ap, ap->exp_runfreq);
 	st_ready(ap);
-
 	st_mutex_unlock();
 
 	free(ra);
@@ -409,6 +405,12 @@ static void do_readmap_mount(struct autofs_point *ap, struct mnt_list *mnts,
 			     me->key);
 			cache_writelock(vmc);
 			valid = cache_lookup_distinct(vmc, me->key);
+			if (!valid) {
+				cache_unlock(vmc);
+				error(ap->logopt,
+				     "failed to find expected existing valid map entry");
+				return;
+			}
 			/* Take over the mount if there is one */
 			valid->ioctlfd = me->ioctlfd;
 			me->ioctlfd = -1;
@@ -486,6 +488,7 @@ static void *do_readmap(void *arg)
 		status = lookup_ghost(ap, ap->path);
 	} else {
 		struct mapent *me;
+		unsigned int append_alarm = !ap->exp_runfreq;
 
 		mnts = tree_make_mnt_tree(_PROC_MOUNTS, "/");
 		pthread_cleanup_push(tree_mnts_cleanup, mnts);
@@ -511,9 +514,19 @@ static void *do_readmap(void *arg)
 			}
 			lookup_prune_one_cache(ap, map->mc, now);
 			pthread_cleanup_pop(1);
+			clear_stale_instances(map);
 			map->stale = 0;
 			map = map->next;
 		}
+
+		/* If the direct mount map was empty at startup no expire
+		 * alarm will have been added. So add it here if there are
+		 * now map entries.
+		 */
+		if (append_alarm && ap->exp_runfreq)
+			alarm_add(ap, ap->exp_runfreq +
+				  rand() % ap->exp_runfreq);
+
 		pthread_cleanup_pop(1);
 		pthread_cleanup_pop(1);
 		pthread_cleanup_pop(1);
@@ -551,7 +564,7 @@ static unsigned int st_readmap(struct autofs_point *ap)
 	pthread_t thid;
 	struct readmap_args *ra;
 	int status;
-	int now = time(NULL);
+	int now = monotonic_time(NULL);
 
 	debug(ap->logopt, "state %d path %s", ap->state, ap->path);
 
@@ -818,7 +831,6 @@ done:
 			list_add(&new->list, head);
 	}
 
-	/* Added task, encourage state machine */
 	signaled = 1;
 	status = pthread_cond_signal(&cond);
 	if (status)
@@ -1129,6 +1141,10 @@ static void *st_queue_handler(void *arg)
 				task = list_entry(p, struct state_queue, list);
 				p = p->next;
 
+				/* Task may have been canceled before it started */
+				if (!task->thid && task->cancel)
+					goto remove;
+
 				if (!task->busy) {
 					/* Start a new task */
 					task->busy = 1;
@@ -1166,9 +1182,9 @@ remove:
 							struct state_queue, pending);
 
 				list_del(&task->list);
+				list_del_init(&next->pending);
 				free(task);
 
-				list_del_init(&next->pending);
 				list_add_tail(&next->list, head);
 				if (p == head)
 					p = head->next;
