@@ -75,6 +75,8 @@ int load_autofs4_module(void);
 #define AUTOFS_SUPER_MAGIC 0x00000187L
 #define SMB_SUPER_MAGIC    0x0000517BL
 #define CIFS_MAGIC_NUMBER  0xFF534D42L
+#define NCP_SUPER_MAGIC    0x0000564CL
+#define NFS_SUPER_MAGIC    0x00006969L
 
 /* This sould be enough for at least 20 host aliases */
 #define HOST_ENT_BUF_SIZE	2048
@@ -281,12 +283,14 @@ int lookup_source_close_ioctlfd(struct autofs_point *ap, const char *key);
 
 #ifdef MODULE_LOOKUP
 int lookup_init(const char *mapfmt, int argc, const char *const *argv, void **context);
+int lookup_reinit(const char *mapfmt, int argc, const char *const *argv, void **context);
 int lookup_read_master(struct master *master, time_t age, void *context);
 int lookup_read_map(struct autofs_point *, time_t, void *context);
 int lookup_mount(struct autofs_point *, const char *, int, void *);
 int lookup_done(void *);
 #endif
 typedef int (*lookup_init_t) (const char *, int, const char *const *, void **);
+typedef int (*lookup_reinit_t) (const char *, int, const char *const *, void **);
 typedef int (*lookup_read_master_t) (struct master *master, time_t, void *);
 typedef int (*lookup_read_map_t) (struct autofs_point *, time_t, void *);
 typedef int (*lookup_mount_t) (struct autofs_point *, const char *, int, void *);
@@ -294,16 +298,21 @@ typedef int (*lookup_done_t) (void *);
 
 struct lookup_mod {
 	lookup_init_t lookup_init;
+	lookup_reinit_t lookup_reinit;
 	lookup_read_master_t lookup_read_master;
 	lookup_read_map_t lookup_read_map;
 	lookup_mount_t lookup_mount;
 	lookup_done_t lookup_done;
+	char *type;
 	void *dlhandle;
 	void *context;
 };
 
-struct lookup_mod *open_lookup(const char *name, const char *err_prefix,
-			       const char *mapfmt, int argc, const char *const *argv);
+int open_lookup(const char *name, const char *err_prefix, const char *mapfmt,
+		int argc, const char *const *argv, struct lookup_mod **lookup);
+int reinit_lookup(struct lookup_mod *mod, const char *name,
+		  const char *err_prefix, const char *mapfmt,
+		  int argc, const char *const *argv);
 int close_lookup(struct lookup_mod *);
 
 /* parse module */
@@ -312,16 +321,19 @@ int close_lookup(struct lookup_mod *);
 
 #ifdef MODULE_PARSE
 int parse_init(int argc, const char *const *argv, void **context);
+int parse_reinit(int argc, const char *const *argv, void **context);
 int parse_mount(struct autofs_point *ap, const char *name,
 		int name_len, const char *mapent, void *context);
 int parse_done(void *);
 #endif
 typedef int (*parse_init_t) (int, const char *const *, void **);
+typedef int (*parse_reinit_t) (int, const char *const *, void **);
 typedef int (*parse_mount_t) (struct autofs_point *, const char *, int, const char *, void *);
 typedef int (*parse_done_t) (void *);
 
 struct parse_mod {
 	parse_init_t parse_init;
+	parse_reinit_t parse_reinit;
 	parse_mount_t parse_mount;
 	parse_done_t parse_done;
 	void *dlhandle;
@@ -330,6 +342,8 @@ struct parse_mod {
 
 struct parse_mod *open_parse(const char *name, const char *err_prefix,
 			     int argc, const char *const *argv);
+int reinit_parse(struct parse_mod *, const char *name,
+		 const char *err_prefix, int argc, const char *const *argv);
 int close_parse(struct parse_mod *);
 
 /* mount module */
@@ -338,17 +352,20 @@ int close_parse(struct parse_mod *);
 
 #ifdef MODULE_MOUNT
 int mount_init(void **context);
+int mount_reinit(void **context);
 int mount_mount(struct autofs_point *ap, const char *root, const char *name, int name_len,
 		const char *what, const char *fstype, const char *options, void *context);
 int mount_done(void *context);
 #endif
 typedef int (*mount_init_t) (void **);
+typedef int (*mount_reinit_t) (void **);
 typedef int (*mount_mount_t) (struct autofs_point *, const char *, const char *, int,
 				const char *, const char *, const char *, void *);
 typedef int (*mount_done_t) (void *);
 
 struct mount_mod {
 	mount_init_t mount_init;
+	mount_reinit_t mount_reinit;
 	mount_mount_t mount_mount;
 	mount_done_t mount_done;
 	void *dlhandle;
@@ -356,6 +373,7 @@ struct mount_mod {
 };
 
 struct mount_mod *open_mount(const char *name, const char *err_prefix);
+int reinit_mount(struct mount_mod *mod, const char *name, const char *err_prefix);
 int close_mount(struct mount_mod *);
 
 /* buffer management */
@@ -413,6 +431,27 @@ struct pending_args {
 };
 
 #ifdef INCLUDE_PENDING_FUNCTIONS
+static void pending_cond_init(void *arg)
+{
+	struct pending_args *mt = (struct pending_args *) arg;
+	pthread_condattr_t condattrs;
+	int status;
+
+	status = pthread_condattr_init(&condattrs);
+	if (status)
+		fatal(status);
+
+	status = pthread_condattr_setclock(&condattrs, CLOCK_MONOTONIC);
+	if (status)
+		fatal(status);
+
+	status = pthread_cond_init(&mt->cond, &condattrs);
+	if (status)
+		fatal(status);
+
+	pthread_condattr_destroy(&condattrs);
+}
+
 static void pending_cond_destroy(void *arg)
 {
 	struct pending_args *mt = (struct pending_args *) arg;
@@ -492,6 +531,7 @@ struct kernel_mod_version {
 struct autofs_point {
 	pthread_t thid;
 	char *path;			/* Mount point name */
+	mode_t mode;			/* Mount point mode */
 	char *pref;			/* amd prefix */
 	int pipefd;			/* File descriptor for pipe */
 	int kpipefd;			/* Kernel end descriptor for pipe */
@@ -565,6 +605,16 @@ do { \
 		fatal(_m_unlock); \
 } while(0)
 
+static inline time_t monotonic_time(time_t *t)
+{
+	struct timespec ts;
+
+	clock_gettime(CLOCK_MONOTONIC, &ts);
+	if (t)
+		*t = (time_t) ts.tv_sec;
+	return (time_t) ts.tv_sec;
+}
+
 /* Expire alarm handling routines */
 int alarm_start_handler(void);
 int alarm_add(struct autofs_point *ap, time_t seconds);
