@@ -203,7 +203,7 @@ int cache_push_mapent(struct mapent *me, char *mapent)
 
 	s->mapent = me->mapent;
 	s->age = me->age;
-	me->mapent = mapent;
+	me->mapent = new;
 
 	if (me->stack)
 		s->next = me->stack;
@@ -733,8 +733,25 @@ int cache_update_offset(struct mapent_cache *mc, const char *mkey, const char *k
 
 	me = cache_lookup_distinct(mc, key);
 	if (me && me->age == age) {
-		if (me == owner || strcmp(me->key, key) == 0)
+		if (me == owner || strcmp(me->key, key) == 0) {
+			char *pent;
+
+			warn(logopt,
+			     "duplcate offset detected for key %s", me->key);
+
+			pent = malloc(strlen(mapent) + 1);
+			if (!pent)
+				warn(logopt,
+				     "map entry not updated: %s", me->mapent);
+			else {
+				if (me->mapent)
+					free(me->mapent);
+				me->mapent = strcpy(pent, mapent);
+				warn(logopt,
+				     "map entry updated with: %s", mapent);
+			}
 			return CHE_DUPLICATE;
+		}
 	}
 
 	ret = cache_update(mc, owner->source, key, mapent, age);
@@ -758,10 +775,14 @@ void cache_update_negative(struct mapent_cache *mc,
 			   struct map_source *ms, const char *key,
 			   time_t timeout)
 {
-	time_t now = time(NULL);
+	time_t now = monotonic_time(NULL);
 	struct mapent *me;
 	int rv = CHE_OK;
 
+	/* Don't update the wildcard */
+	if (strlen(key) == 1 && *key == '*')
+		return;
+
 	me = cache_lookup_distinct(mc, key);
 	if (me)
 		rv = cache_push_mapent(me, NULL);
@@ -1162,7 +1183,6 @@ struct mapent *cache_enumerate(struct mapent_cache *mc, struct mapent *me)
  * Get each offset from list head under prefix.
  * Maintain traversal current position in pos for subsequent calls. 
  * Return each offset into offset.
- * TODO: length check on offset.
  */
 /* cache must be read locked by caller */
 char *cache_get_offset(const char *prefix, char *offset, int start,
@@ -1191,6 +1211,9 @@ char *cache_get_offset(const char *prefix, char *offset, int start,
 			continue;
 
 		if (!strncmp(prefix, offset_start, plen)) {
+			struct mapent *np = NULL;
+			char pe[PATH_MAX + 1];
+
 			/* "/" doesn't count for root offset */
 			if (plen == 1)
 				pstart = &offset_start[plen - 1];
@@ -1203,7 +1226,24 @@ char *cache_get_offset(const char *prefix, char *offset, int start,
 
 			/* get next offset */
 			pend = pstart;
-			while (*pend++) ;
+			while (*pend++) {
+				size_t nest_pt_offset;
+
+				if (*pend != '/')
+					continue;
+
+				nest_pt_offset = start + pend - pstart;
+				if (plen > 1)
+					nest_pt_offset += plen;
+				strcpy(pe, this->key);
+				pe[nest_pt_offset] = '\0';
+
+				np = cache_lookup_distinct(this->mc, pe);
+				if (np)
+					break;
+			}
+			if (np)
+				continue;
 			len = pend - pstart - 1;
 			strncpy(offset, pstart, len);
 			offset[len] ='\0';
@@ -1236,7 +1276,9 @@ char *cache_get_offset(const char *prefix, char *offset, int start,
 			break;
 
 		/* compare offset */
-		if (pstart[len] != '/' || strncmp(offset, pstart, len))
+		if (pstart[len] != '/' ||
+		    strlen(pstart) != len ||
+		    strncmp(offset, pstart, len))
 			break;
 
 		*pos = next;
