@@ -26,11 +26,13 @@
 #include <sys/vfs.h>
 #include <pwd.h>
 #include <grp.h>
+#include <libgen.h>
 
 #include "automount.h"
 
 #define MAX_OPTIONS_LEN		80
 #define MAX_MNT_NAME_LEN	30
+#define MAX_ENV_NAME		15
 
 #define EBUFSIZ 1024
 
@@ -46,6 +48,18 @@ static const char mnt_name_template[]      = "automount(pid%u)";
 static struct kernel_mod_version kver = {0, 0};
 static const char kver_options_template[]  = "fd=%d,pgrp=%u,minproto=3,maxproto=5";
 
+#define EXT_MOUNTS_HASH_SIZE    50
+
+struct ext_mount {
+	char *mountpoint;
+	unsigned int umount;
+	struct list_head mount;
+	struct list_head mounts;
+};
+static struct list_head ext_mounts_hash[EXT_MOUNTS_HASH_SIZE];
+static unsigned int ext_mounts_hash_init_done = 0;
+static pthread_mutex_t ext_mount_hash_mutex = PTHREAD_MUTEX_INITIALIZER;
+
 unsigned int linux_version_code(void)
 {
 	struct utsname my_utsname;
@@ -77,7 +91,7 @@ unsigned int linux_version_code(void)
 
 unsigned int query_kproto_ver(void)
 {
-	struct ioctl_ops *ops = get_ioctl_ops();
+	struct ioctl_ops *ops;
 	char dir[] = "/tmp/autoXXXXXX", *t_dir;
 	char options[MAX_OPTIONS_LEN + 1];
 	pid_t pgrp = getpgrp();
@@ -118,10 +132,19 @@ unsigned int query_kproto_ver(void)
 		return 0;
 	}
 
+	ops = get_ioctl_ops();
+	if (!ops) {
+		umount(t_dir);
+		close(pipefd[0]);
+		rmdir(t_dir);
+		return 0;
+	}
+
 	ops->open(LOGOPT_NONE, &ioctlfd, st.st_dev, t_dir);
 	if (ioctlfd == -1) {
 		umount(t_dir);
 		close(pipefd[0]);
+		close_ioctl_ctl();
 		rmdir(t_dir);
 		return 0;
 	}
@@ -133,6 +156,7 @@ unsigned int query_kproto_ver(void)
 		ops->close(LOGOPT_NONE, ioctlfd);
 		umount(t_dir);
 		close(pipefd[0]);
+		close_ioctl_ctl();
 		rmdir(t_dir);
 		return 0;
 	}
@@ -142,6 +166,7 @@ unsigned int query_kproto_ver(void)
 		ops->close(LOGOPT_NONE, ioctlfd);
 		umount(t_dir);
 		close(pipefd[0]);
+		close_ioctl_ctl();
 		rmdir(t_dir);
 		return 0;
 	}
@@ -149,6 +174,7 @@ unsigned int query_kproto_ver(void)
 	ops->close(LOGOPT_NONE, ioctlfd);
 	umount(t_dir);
 	close(pipefd[0]);
+	close_ioctl_ctl();
 	rmdir(t_dir);
 
 	return 1;
@@ -303,7 +329,61 @@ int check_nfs_mount_version(struct nfs_mount_vers *vers,
 }
 #endif
 
-struct substvar *addstdenv(struct substvar *sv)
+static char *set_env_name(const char *prefix, const char *name, char *buf)
+{
+	size_t len;
+
+	len = strlen(name);
+	if (prefix)
+		len += strlen(prefix);
+	len++;
+
+	if (len > MAX_ENV_NAME)
+		return NULL;
+
+	if (!prefix)
+		strcpy(buf, name);
+	else {
+		strcpy(buf, prefix);
+		strcat(buf, name);
+	}
+	return buf;
+}
+
+static struct substvar *do_macro_addvar(struct substvar *list,
+					const char *prefix,
+					const char *name,
+					const char *val)
+{
+	char buf[MAX_ENV_NAME + 1];
+	char *new;
+	size_t len;
+
+	new = set_env_name(prefix, name, buf);
+	if (new) {
+		len = strlen(new);
+		list = macro_addvar(list, new, len, val);
+	}
+	return list;
+}
+
+static struct substvar *do_macro_removevar(struct substvar *list,
+					   const char *prefix,
+					   const char *name)
+{
+	char buf[MAX_ENV_NAME + 1];
+	char *new;
+	size_t len;
+
+	new = set_env_name(prefix, name, buf);
+	if (new) {
+		len = strlen(new);
+		list = macro_removevar(list, new, len);
+	}
+	return list;
+}
+
+struct substvar *addstdenv(struct substvar *sv, const char *prefix)
 {
 	struct substvar *list = sv;
 	struct thread_stdenv_vars *tsv;
@@ -318,14 +398,14 @@ struct substvar *addstdenv(struct substvar *sv)
 		num = (long) tsv->uid;
 		ret = sprintf(numbuf, "%ld", num);
 		if (ret > 0)
-			list = macro_addvar(list, "UID", 3, numbuf);
+			list = do_macro_addvar(list, prefix, "UID", numbuf);
 		num = (long) tsv->gid;
 		ret = sprintf(numbuf, "%ld", num);
 		if (ret > 0)
-			list = macro_addvar(list, "GID", 3, numbuf);
-		list = macro_addvar(list, "USER", 4, tsv->user);
-		list = macro_addvar(list, "GROUP", 5, tsv->group);
-		list = macro_addvar(list, "HOME", 4, tsv->home);
+			list = do_macro_addvar(list, prefix, "GID", numbuf);
+		list = do_macro_addvar(list, prefix, "USER", tsv->user);
+		list = do_macro_addvar(list, prefix, "GROUP", tsv->group);
+		list = do_macro_addvar(list, prefix, "HOME", tsv->home);
 		mv = macro_findvar(list, "HOST", 4);
 		if (mv) {
 			char *shost = strdup(mv->val);
@@ -333,7 +413,8 @@ struct substvar *addstdenv(struct substvar *sv)
 				char *dot = strchr(shost, '.');
 				if (dot)
 					*dot = '\0';
-				list = macro_addvar(list, "SHOST", 5, shost);
+				list = do_macro_addvar(list,
+						       prefix, "SHOST", shost);
 				free(shost);
 			}
 		}
@@ -341,19 +422,176 @@ struct substvar *addstdenv(struct substvar *sv)
 	return list;
 }
 
-struct substvar *removestdenv(struct substvar *sv)
+struct substvar *removestdenv(struct substvar *sv, const char *prefix)
 {
 	struct substvar *list = sv;
 
-	list = macro_removevar(list, "UID", 3);
-	list = macro_removevar(list, "USER", 4);
-	list = macro_removevar(list, "HOME", 4);
-	list = macro_removevar(list, "GID", 3);
-	list = macro_removevar(list, "GROUP", 5);
-	list = macro_removevar(list, "SHOST", 5);
+	list = do_macro_removevar(list, prefix, "UID");
+	list = do_macro_removevar(list, prefix, "USER");
+	list = do_macro_removevar(list, prefix, "HOME");
+	list = do_macro_removevar(list, prefix, "GID");
+	list = do_macro_removevar(list, prefix, "GROUP");
+	list = do_macro_removevar(list, prefix, "SHOST");
 	return list;
 }
 
+void add_std_amd_vars(struct substvar *sv)
+{
+	char *tmp;
+
+	tmp = conf_amd_get_arch();
+	if (tmp) {
+		macro_global_addvar("arch", 4, tmp);
+		free(tmp);
+	}
+
+	tmp = conf_amd_get_karch();
+	if (tmp) {
+		macro_global_addvar("karch", 5, tmp);
+		free(tmp);
+	}
+
+	tmp = conf_amd_get_os();
+	if (tmp) {
+		macro_global_addvar("os", 2, tmp);
+		free(tmp);
+	}
+
+	tmp = conf_amd_get_full_os();
+	if (tmp) {
+		macro_global_addvar("full_os", 7, tmp);
+		free(tmp);
+	}
+
+	tmp = conf_amd_get_os_ver();
+	if (tmp) {
+		macro_global_addvar("osver", 5, tmp);
+		free(tmp);
+	}
+
+	tmp = conf_amd_get_vendor();
+	if (tmp) {
+		macro_global_addvar("vendor", 6, tmp);
+		free(tmp);
+	}
+
+	/* Umm ... HP_UX cluster name, probably not used */
+	tmp = conf_amd_get_cluster();
+	if (tmp) {
+		macro_global_addvar("cluster", 7, tmp);
+		free(tmp);
+	} else {
+		const struct substvar *v = macro_findvar(sv, "domain", 4);
+		if (v && *v->val) {
+			tmp = strdup(v->val);
+			if (tmp)
+				macro_global_addvar("cluster", 7, tmp);
+		}
+	}
+
+	tmp = conf_amd_get_auto_dir();
+	if (tmp) {
+		macro_global_addvar("autodir", 7, tmp);
+		free(tmp);
+	}
+
+	return;
+}
+
+void remove_std_amd_vars(void)
+{
+	macro_global_removevar("autodir", 7);
+	macro_global_removevar("cluster", 7);
+	macro_global_removevar("vendor", 6);
+	macro_global_removevar("osver", 5);
+	macro_global_removevar("full_os", 7);
+	macro_global_removevar("os", 2);
+	macro_global_removevar("karch", 5);
+	macro_global_removevar("arch", 4);
+	return;
+ }
+
+struct amd_entry *new_amd_entry(const struct substvar *sv)
+{
+	struct amd_entry *new;
+	const struct substvar *v;
+	char *path;
+
+	v = macro_findvar(sv, "path", 4);
+	if (!v)
+		return NULL;
+
+	path = strdup(v->val);
+	if (!path)
+		return NULL;
+
+	new = malloc(sizeof(struct amd_entry));
+	if (!new) {
+		free(path);
+		return NULL;
+	}
+
+	memset(new, 0, sizeof(*new));
+	new->path = path;
+	INIT_LIST_HEAD(&new->list);
+	INIT_LIST_HEAD(&new->entries);
+	INIT_LIST_HEAD(&new->ext_mount);
+
+	return new;
+}
+
+void clear_amd_entry(struct amd_entry *entry)
+{
+	if (!entry)
+		return;
+	if (entry->path)
+		free(entry->path);
+	if (entry->map_type)
+		free(entry->map_type);
+	if (entry->pref)
+		free(entry->pref);
+	if (entry->fs)
+		free(entry->fs);
+	if (entry->rhost)
+		free(entry->rhost);
+	if (entry->rfs)
+		free(entry->rfs);
+	if (entry->opts)
+		free(entry->opts);
+	if (entry->addopts)
+		free(entry->addopts);
+	if (entry->remopts)
+		free(entry->remopts);
+	if (entry->sublink)
+		free(entry->sublink);
+	if (entry->selector)
+		free_selector(entry->selector);
+	return;
+}
+
+void free_amd_entry(struct amd_entry *entry)
+{
+	clear_amd_entry(entry);
+	free(entry);
+	return;
+}
+
+void free_amd_entry_list(struct list_head *entries)
+{
+	if (!list_empty(entries)) {
+		struct list_head *head = entries;
+		struct amd_entry *this;
+		struct list_head *p;
+
+		p = head->next;
+		while (p != head) {
+			this = list_entry(p, struct amd_entry, list);
+			p = p->next;
+			free_amd_entry(this);
+		}
+	}
+}
+
 /*
  * Make common autofs mount options string
  */
@@ -422,6 +660,129 @@ char *make_mnt_name_string(char *path)
 	return mnt_name;
 }
 
+static void ext_mounts_hash_init(void)
+{
+	int i;
+	for (i = 0; i < EXT_MOUNTS_HASH_SIZE; i++)
+		INIT_LIST_HEAD(&ext_mounts_hash[i]);
+	ext_mounts_hash_init_done = 1;
+}
+
+static struct ext_mount *ext_mount_lookup(const char *mountpoint)
+{
+	u_int32_t hval = hash(mountpoint, EXT_MOUNTS_HASH_SIZE);
+	struct list_head *p, *head;
+
+	if (!ext_mounts_hash_init_done)
+		ext_mounts_hash_init();
+
+	if (list_empty(&ext_mounts_hash[hval]))
+		return NULL;
+
+	head = &ext_mounts_hash[hval];
+	list_for_each(p, head) {
+		struct ext_mount *this = list_entry(p, struct ext_mount, mount);
+		if (!strcmp(this->mountpoint, mountpoint))
+			return this;
+	}
+	return NULL;
+}
+
+int ext_mount_add(struct list_head *entry, const char *path, unsigned int umount)
+{
+	struct ext_mount *em;
+	char *auto_dir;
+	u_int32_t hval;
+	int ret = 0;
+
+	/* Not a mount in the external mount directory */
+	auto_dir = conf_amd_get_auto_dir();
+	if (strncmp(path, auto_dir, strlen(auto_dir))) {
+		free(auto_dir);
+		return 0;
+	}
+	free(auto_dir);
+
+	pthread_mutex_lock(&ext_mount_hash_mutex);
+
+	em = ext_mount_lookup(path);
+	if (em) {
+		struct list_head *p, *head;
+		head = &em->mounts;
+		list_for_each(p, head) {
+			if (p == entry)
+				goto done;
+		}
+		list_add_tail(entry, &em->mounts);
+		ret = 1;
+		goto done;
+	}
+
+	em = malloc(sizeof(struct ext_mount));
+	if (!em) {
+		ret = -1;
+		goto done;
+	}
+
+	em->mountpoint = strdup(path);
+	if (!em->mountpoint) {
+		free(em);
+		ret = -1;
+		goto done;
+	}
+	em->umount = umount;
+	INIT_LIST_HEAD(&em->mount);
+	INIT_LIST_HEAD(&em->mounts);
+
+	hval = hash(path, EXT_MOUNTS_HASH_SIZE);
+	list_add_tail(&em->mount, &ext_mounts_hash[hval]);
+
+	list_add_tail(entry, &em->mounts);
+
+	ret = 1;
+done:
+	pthread_mutex_unlock(&ext_mount_hash_mutex);
+	return ret;
+}
+
+int ext_mount_remove(struct list_head *entry, const char *path)
+{
+	struct ext_mount *em;
+	char *auto_dir;
+	int ret = 0;
+
+	/* Not a mount in the external mount directory */
+	auto_dir = conf_amd_get_auto_dir();
+	if (strncmp(path, auto_dir, strlen(auto_dir))) {
+		free(auto_dir);
+		return 0;
+	}
+	free(auto_dir);
+
+	pthread_mutex_lock(&ext_mount_hash_mutex);
+
+	em = ext_mount_lookup(path);
+	if (!em)
+		goto done;
+
+	list_del_init(entry);
+
+	if (!list_empty(&em->mounts))
+		goto done;
+	else {
+		list_del_init(&em->mount);
+		if (em->umount)
+			ret = 1;
+		if (list_empty(&em->mount)) {
+			free(em->mountpoint);
+			free(em);
+		}
+	}
+done:
+	pthread_mutex_unlock(&ext_mount_hash_mutex);
+	return ret;
+}
+
 /*
  * Get list of mounts under path in longest->shortest order
  */
@@ -580,51 +941,6 @@ void free_mnt_list(struct mnt_list *list)
 	}
 }
 
-int contained_in_local_fs(const char *path)
-{
-	struct mnt_list *mnts, *this;
-	size_t pathlen = strlen(path);
-	int ret;
-
-	if (!path || !pathlen || pathlen > PATH_MAX)
-		return 0;
-
-	mnts = get_mnt_list(_PATH_MOUNTED, "/", 1);
-	if (!mnts)
-		return 0;
-
-	ret = 0;
-
-	for (this = mnts; this != NULL; this = this->next) {
-		size_t len = strlen(this->path);
-
-		if (!strncmp(path, this->path, len)) {
-			if (len > 1 && pathlen > len && path[len] != '/')
-				continue;
-			else if (len == 1 && this->path[0] == '/') {
-				/*
-				 * always return true on rootfs, we don't
-				 * want to break diskless clients.
-				 */
-				ret = 1;
-			} else if (this->fs_name[0] == '/') {
-				if (strlen(this->fs_name) > 1) {
-					if (this->fs_name[1] != '/')
-						ret = 1;
-				} else
-					ret = 1;
-			} else if (!strncmp("LABEL=", this->fs_name, 6) ||
-				   !strncmp("UUID=", this->fs_name, 5))
-				ret = 1;
-			break;
-		}
-	}
-
-	free_mnt_list(mnts);
-
-	return ret;
-}
-
 static int table_is_mounted(const char *table, const char *path, unsigned int type)
 {
 	struct mntent *mnt;
@@ -671,12 +987,19 @@ static int table_is_mounted(const char *table, const char *path, unsigned int ty
 	return ret;
 }
 
-static int ioctl_is_mounted(const char *path, unsigned int type)
+static int ioctl_is_mounted(const char *table, const char *path, unsigned int type)
 {
 	struct ioctl_ops *ops = get_ioctl_ops();
 	unsigned int mounted;
+	int ret;
+
+	/* If the ioctl fails fall back to the potentially resource
+	 * intensive mount table check.
+	 */
+	ret = ops->ismountpoint(LOGOPT_NONE, -1, path, &mounted);
+	if (ret == -1)
+		return table_is_mounted(table, path, type);
 
-	ops->ismountpoint(LOGOPT_NONE, -1, path, &mounted);
 	if (mounted) {
 		switch (type) {
 		case MNTS_ALL:
@@ -695,7 +1018,7 @@ int is_mounted(const char *table, const char *path, unsigned int type)
 	struct ioctl_ops *ops = get_ioctl_ops();
 
 	if (ops->ismountpoint)
-		return ioctl_is_mounted(path, type);
+		return ioctl_is_mounted(table, path, type);
 	else
 		return table_is_mounted(table, path, type);
 }
@@ -729,100 +1052,6 @@ int has_fstab_option(const char *opt)
 	return ret;
 }
 
-char *get_offset(const char *prefix, char *offset,
-		 struct list_head *head, struct list_head **pos)
-{
-	struct list_head *next;
-	struct mnt_list *this;
-	size_t plen = strlen(prefix);
-	size_t len = 0;
-
-	*offset = '\0';
-	next = *pos ? (*pos)->next : head->next;
-	while (next != head) {
-		char *pstart, *pend;
-
-		this = list_entry(next, struct mnt_list, ordered);
-		*pos = next;
-		next = next->next;
-
-		if (strlen(this->path) <= plen)
-			continue;
-
-		if (!strncmp(prefix, this->path, plen)) {
-			pstart = &this->path[plen];
-
-			/* not part of this sub-tree */
-			if (*pstart != '/')
-				continue;
-
-			/* get next offset */
-			pend = pstart;
-			while (*pend++) ;
-			len = pend - pstart - 1;
-			strncpy(offset, pstart, len);
-			offset[len] ='\0';
-			break;
-		}
-	}
-
-	while (next != head) {
-		char *pstart;
-
-		this = list_entry(next, struct mnt_list, ordered);
-
-		if (strlen(this->path) <= plen + len)
-			break;
-
-		pstart = &this->path[plen];
-
-		/* not part of this sub-tree */
-		if (*pstart != '/')
-			break;
-
-		/* new offset */
-		if (!*(pstart + len + 1))
-			break;
-
-		/* compare next offset */
-		if (pstart[len] != '/' || strncmp(offset, pstart, len))
-			break;
-
-		*pos = next;
-		next = next->next;
-	}
-
-	return *offset ? offset : NULL;
-}
-
-void add_ordered_list(struct mnt_list *ent, struct list_head *head)
-{
-	struct list_head *p;
-	struct mnt_list *this;
-
-	list_for_each(p, head) {
-		size_t tlen;
-		int eq;
-
-		this = list_entry(p, struct mnt_list, ordered);
-		tlen = strlen(this->path);
-
-		eq = strncmp(this->path, ent->path, tlen);
-		if (!eq && tlen == strlen(ent->path))
-			return;
-
-		if (eq > 0) {
-			INIT_LIST_HEAD(&ent->ordered);
-			list_add_tail(&ent->ordered, p);
-			return;
-		}
-	}
-	INIT_LIST_HEAD(&ent->ordered);
-	list_add_tail(&ent->ordered, p);
-
-	return;
-}
-
 /*
  * Since we have to look at the entire mount tree for direct
  * mounts (all mounts under "/") and we may have a large number
@@ -922,7 +1151,6 @@ struct mnt_list *tree_make_mnt_tree(const char *table, const char *path)
 		INIT_LIST_HEAD(&ent->list);
 		INIT_LIST_HEAD(&ent->entries);
 		INIT_LIST_HEAD(&ent->sublist);
-		INIT_LIST_HEAD(&ent->ordered);
 
 		ent->path = malloc(len + 1);
 		if (!ent->path) {
@@ -1173,7 +1401,7 @@ int tree_is_mounted(struct mnt_list *mnts, const char *path, unsigned int type)
 	int mounted = 0;
 
 	if (ops->ismountpoint)
-		return ioctl_is_mounted(path, type);
+		return ioctl_is_mounted(_PROC_MOUNTS, path, type);
 
 	INIT_LIST_HEAD(&list);
 
@@ -1860,11 +2088,12 @@ int umount_multi_triggers(struct autofs_point *ap, struct mapent *me, char *root
 		 */
 		if (is_mounted(_PATH_MOUNTED, root, MNTS_REAL)) {
 			info(ap->logopt, "unmounting dir = %s", root);
-			if (umount_ent(ap, root)) {
+			if (umount_ent(ap, root) &&
+			    is_mounted(_PATH_MOUNTED, root, MNTS_REAL)) {
 				if (mount_multi_triggers(ap, me, root, strlen(root), "/") < 0)
 					warn(ap->logopt,
 					     "failed to remount offset triggers");
-				return left++;
+				return ++left;
 			}
 		}
 
@@ -1932,7 +2161,7 @@ int clean_stale_multi_triggers(struct autofs_point *ap,
 		/* Check for and umount stale subtree offsets */
 		oe_base = oe->key + strlen(root);
 		ret = clean_stale_multi_triggers(ap, oe, root, oe_base);
-		left =+ ret;
+		left += ret;
 		if (ret)
 			continue;
 
@@ -1961,7 +2190,8 @@ int clean_stale_multi_triggers(struct autofs_point *ap,
 		 */
 		if (oe->ioctlfd != -1 ||
 		    is_mounted(_PROC_MOUNTS, oe->key, MNTS_REAL)) {
-			if (umount_ent(ap, oe->key)) {
+			if (umount_ent(ap, oe->key) &&
+			    is_mounted(_PROC_MOUNTS, oe->key, MNTS_REAL)) {
 				debug(ap->logopt,
 				      "offset %s has active mount, invalidate",
 				      oe->key);
@@ -1977,7 +2207,7 @@ int clean_stale_multi_triggers(struct autofs_point *ap,
 		key = strdup(oe->key);
 		if (!key) {
 	                char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-        	        error(ap->logopt, "malloc: %s", estr);
+		        error(ap->logopt, "malloc: %s", estr);
 			left++;
 			continue;
 		}
