@@ -33,16 +33,6 @@
 #include <pthread.h>
 #include <poll.h>
 
-#ifdef WITH_LIBTIRPC
-#undef auth_destroy
-#define auth_destroy(auth)                                              \
-                do {                                                    \
-                        int refs;                                       \
-                        if ((refs = auth_put((auth))) == 0)             \
-                                ((*((auth)->ah_ops->ah_destroy))(auth));\
-                } while (0)
-#endif
-
 #ifdef WITH_LIBTIRPC
 const rpcprog_t rpcb_prog = RPCBPROG;
 const rpcvers_t rpcb_version = RPCBVERS;
@@ -77,7 +67,6 @@ static const rpcvers_t mount_vers[] = {
 };
 
 static int connect_nb(int, struct sockaddr *, socklen_t, struct timeval *);
-inline void dump_core(void);
 
 /*
  *  Perform a non-blocking connect on the socket fd.
@@ -1078,12 +1067,15 @@ int rpc_ping(const char *host, long seconds, long micros, unsigned int option)
 	return status;
 }
 
-double elapsed(struct timeval start, struct timeval end)
+double monotonic_elapsed(struct timespec start, struct timespec end)
 {
 	double t1, t2;
-	t1 =  (double)start.tv_sec + (double)start.tv_usec/(1000*1000);
-	t2 =  (double)end.tv_sec + (double)end.tv_usec/(1000*1000);
-	return t2-t1;
+
+	t1 =  (double) start.tv_sec +
+		(double) (start.tv_nsec/(1000*1000*1000));
+	t2 =  (double) end.tv_sec +
+		(double) (end.tv_nsec/(1000*1000*1000));
+	return t2 - t1;
 }
 
 int rpc_time(const char *host,
@@ -1092,19 +1084,18 @@ int rpc_time(const char *host,
 {
 	int status;
 	double taken;
-	struct timeval start, end;
-	struct timezone tz;
+	struct timespec start, end;
 	int proto = (ping_proto & RPC_PING_UDP) ? IPPROTO_UDP : IPPROTO_TCP;
 	unsigned long vers = ping_vers;
 
-	gettimeofday(&start, &tz);
+	clock_gettime(CLOCK_MONOTONIC, &start);
 	status = __rpc_ping(host, vers, proto, seconds, micros, option);
-	gettimeofday(&end, &tz);
+	clock_gettime(CLOCK_MONOTONIC, &end);
 
 	if (status == RPC_PING_FAIL || status < 0)
 		return status;
 
-	taken = elapsed(start, end);
+	taken = monotonic_elapsed(start, end);
 
 	if (result != NULL)
 		*result = taken;
