@@ -214,35 +214,45 @@ int bind_ldap_simple(unsigned logopt, LDAP *ldap, const char *uri, struct lookup
 	return 0;
 }
 
-int __unbind_ldap_connection(unsigned logopt, LDAP *ldap, struct lookup_context *ctxt)
+int __unbind_ldap_connection(unsigned logopt,
+			     struct ldap_conn *conn,
+			     struct lookup_context *ctxt)
 {
-	int rv;
+	int rv = LDAP_SUCCESS;
 
 	if (ctxt->use_tls == LDAP_TLS_RELEASE)
 		ctxt->use_tls = LDAP_TLS_INIT;
 #ifdef WITH_SASL
-	autofs_sasl_unbind(ctxt);
+	if (ctxt->auth_required & LDAP_NEED_AUTH)
+		autofs_sasl_unbind(conn, ctxt);
+	/* No, sasl_dispose does not release the ldap connection
+	 * unless it's using sasl EXTERNAL
+	 */
 #endif
-
-	rv = ldap_unbind_ext(ldap, NULL, NULL);
+	if (conn->ldap) {
+		rv = ldap_unbind_ext(conn->ldap, NULL, NULL);
+		conn->ldap = NULL;
+	}
 	if (rv != LDAP_SUCCESS)
 		error(logopt, "unbind failed: %s", ldap_err2string(rv));
 
 	return rv;
 }
 
-int unbind_ldap_connection(unsigned logopt, LDAP *ldap, struct lookup_context *ctxt)
+int unbind_ldap_connection(unsigned logopt,
+			   struct ldap_conn *conn,
+			   struct lookup_context *ctxt)
 {
 	int rv;
 
 	ldapinit_mutex_lock();
-	rv = __unbind_ldap_connection(logopt, ldap, ctxt);
+	rv = __unbind_ldap_connection(logopt, conn, ctxt);
 	ldapinit_mutex_unlock();
 
 	return rv;
 }
 
-LDAP *__init_ldap_connection(unsigned logopt, const char *uri, struct lookup_context *ctxt)
+LDAP *init_ldap_connection(unsigned logopt, const char *uri, struct lookup_context *ctxt)
 {
 	LDAP *ldap = NULL;
 	struct timeval timeout     = { ctxt->timeout, 0 };
@@ -302,7 +312,7 @@ LDAP *__init_ldap_connection(unsigned logopt, const char *uri, struct lookup_con
 
 		rv = ldap_start_tls_s(ldap, NULL, NULL);
 		if (rv != LDAP_SUCCESS) {
-			__unbind_ldap_connection(logopt, ldap, ctxt);
+			ldap_unbind_ext(ldap, NULL, NULL);
 			if (ctxt->tls_required) {
 				error(logopt, MODPREFIX
 				      "TLS required but START_TLS failed: %s",
@@ -310,7 +320,7 @@ LDAP *__init_ldap_connection(unsigned logopt, const char *uri, struct lookup_con
 				return NULL;
 			}
 			ctxt->use_tls = LDAP_TLS_DONT_USE;
-			ldap = __init_ldap_connection(logopt, uri, ctxt);
+			ldap = init_ldap_connection(logopt, uri, ctxt);
 			if (ldap)
 				ctxt->use_tls = LDAP_TLS_INIT;
 			return ldap;
@@ -321,17 +331,6 @@ LDAP *__init_ldap_connection(unsigned logopt, const char *uri, struct lookup_con
 	return ldap;
 }
 
-LDAP *init_ldap_connection(unsigned logopt, const char *uri, struct lookup_context *ctxt)
-{
-	LDAP *ldap;
-
-	ldapinit_mutex_lock();
-	ldap = __init_ldap_connection(logopt, uri, ctxt);
-	ldapinit_mutex_unlock();
-
-	return ldap;
-}
-
 static int get_query_dn(unsigned logopt, LDAP *ldap, struct lookup_context *ctxt, const char *class, const char *key)
 {
 	char buf[MAX_ERR_BUF];
@@ -571,33 +570,32 @@ static int find_query_dn(unsigned logopt, LDAP *ldap, struct lookup_context *ctx
 	return 0;
 }
 
-static int do_bind(unsigned logopt, LDAP *ldap, const char *uri, struct lookup_context *ctxt)
+static int do_bind(unsigned logopt, struct ldap_conn *conn,
+		   const char *uri, struct lookup_context *ctxt)
 {
 	char *host = NULL, *nhost;
-	int rv, need_base = 1;
+	int rv;
 
 #ifdef WITH_SASL
 	debug(logopt, MODPREFIX "auth_required: %d, sasl_mech %s",
 	      ctxt->auth_required, ctxt->sasl_mech);
 
 	if (ctxt->auth_required & LDAP_NEED_AUTH) {
-		ldapinit_mutex_lock();
-		rv = autofs_sasl_bind(logopt, ldap, ctxt);
-		ldapinit_mutex_unlock();
+		rv = autofs_sasl_bind(logopt, conn, ctxt);
 		debug(logopt, MODPREFIX "autofs_sasl_bind returned %d", rv);
 	} else {
-		rv = bind_ldap_simple(logopt, ldap, uri, ctxt);
+		rv = bind_ldap_simple(logopt, conn->ldap, uri, ctxt);
 		debug(logopt, MODPREFIX "ldap simple bind returned %d", rv);
 	}
 #else
-	rv = bind_ldap_simple(logopt, ldap, uri, ctxt);
+	rv = bind_ldap_simple(logopt, conn->ldap, uri, ctxt);
 	debug(logopt, MODPREFIX "ldap simple bind returned %d", rv);
 #endif
 
 	if (rv != 0)
 		return 0;
 
-	rv = ldap_get_option(ldap, LDAP_OPT_HOST_NAME, &host);
+	rv = ldap_get_option(conn->ldap, LDAP_OPT_HOST_NAME, &host);
         if (rv != LDAP_SUCCESS || !host) {
 		debug(logopt, "failed to get hostname for connection");
 		return 0;
@@ -610,6 +608,7 @@ static int do_bind(unsigned logopt, LDAP *ldap, const char *uri, struct lookup_c
 	}
 	ldap_memfree(host);
 
+	uris_mutex_lock(ctxt);
 	if (!ctxt->cur_host) {
 		ctxt->cur_host = nhost;
 		if (!(ctxt->format & MAP_FLAG_FORMAT_AMD)) {
@@ -618,44 +617,23 @@ static int do_bind(unsigned logopt, LDAP *ldap, const char *uri, struct lookup_c
 		}
 	} else {
 		/* If connection host has changed update */
-		if (strcmp(ctxt->cur_host, nhost)) {
+		if (!strcmp(ctxt->cur_host, nhost))
+			free(nhost);
+		else {
 			free(ctxt->cur_host);
 			ctxt->cur_host = nhost;
-		} else {
-			free(nhost);
-			need_base = 0;
-		}
-	}
-
-	if (ctxt->schema && ctxt->qdn && !need_base)
-		return 1;
-
-	/*
-	 * If the schema isn't defined in the configuration then check for
-	 * presence of a map dn with a the common schema. Then calculate the
-	 * base dn for searches.
-	 */
-	if (!ctxt->schema) {
-		if (!find_query_dn(logopt, ldap, ctxt)) {
-			warn(logopt,
-			      MODPREFIX "failed to find valid query dn");
-			return 0;
-		}
-	} else if (!(ctxt->format & MAP_FLAG_FORMAT_AMD)) {
-		const char *class = ctxt->schema->map_class;
-		const char *key = ctxt->schema->map_attr;
-		if (!get_query_dn(logopt, ldap, ctxt, class, key)) {
-			error(logopt, MODPREFIX "failed to get query dn");
-			return 0;
 		}
 	}
+	uris_mutex_unlock(ctxt);
 
 	return 1;
 }
 
-static LDAP *do_connect(unsigned logopt, const char *uri, struct lookup_context *ctxt)
+static int do_connect(unsigned logopt, struct ldap_conn *conn,
+		      const char *uri, struct lookup_context *ctxt)
 {
-	LDAP *ldap;
+	char *cur_host = NULL;
+	int ret = NSS_STATUS_SUCCESS;
 
 #ifdef WITH_SASL
 	if (ctxt->extern_cert && ctxt->extern_key) {
@@ -664,19 +642,64 @@ static LDAP *do_connect(unsigned logopt, const char *uri, struct lookup_context
 	}
 #endif
 
-	ldap = init_ldap_connection(logopt, uri, ctxt);
-	if (ldap) {
-		if (!do_bind(logopt, ldap, uri, ctxt)) {
-			unbind_ldap_connection(logopt, ldap, ctxt);
-			ldap = NULL;
+	conn->ldap = init_ldap_connection(logopt, uri, ctxt);
+	if (!conn->ldap) {
+		ret = NSS_STATUS_UNAVAIL;
+		goto out;
+	}
+
+	uris_mutex_lock(ctxt);
+	if (ctxt->cur_host)
+		cur_host = ctxt->cur_host;
+	uris_mutex_unlock(ctxt);
+
+	if (!do_bind(logopt, conn, uri, ctxt)) {
+		__unbind_ldap_connection(logopt, conn, ctxt);
+		ret = NSS_STATUS_UNAVAIL;
+		goto out;
+	}
+
+	/* If the lookup schema and the query dn are set and the
+	 * ldap host hasn't changed return.
+	 */
+	uris_mutex_lock(ctxt);
+	if (ctxt->schema && ctxt->qdn && (cur_host == ctxt->cur_host)) {
+		uris_mutex_unlock(ctxt);
+		goto out;
+	}
+	uris_mutex_unlock(ctxt);
+
+	/*
+	 * If the schema isn't defined in the configuration then check for
+	 * presence of a map dn with a the common schema. Then calculate the
+	 * base dn for searches.
+	 */
+	if (!ctxt->schema) {
+		if (!find_query_dn(logopt, conn->ldap, ctxt)) {
+			__unbind_ldap_connection(logopt, conn, ctxt);
+			ret = NSS_STATUS_NOTFOUND;
+			warn(logopt,
+			      MODPREFIX "failed to find valid query dn");
+			goto out;
+		}
+	} else if (!(ctxt->format & MAP_FLAG_FORMAT_AMD)) {
+		const char *class = ctxt->schema->map_class;
+		const char *key = ctxt->schema->map_attr;
+		if (!get_query_dn(logopt, conn->ldap, ctxt, class, key)) {
+			__unbind_ldap_connection(logopt, conn, ctxt);
+			ret = NSS_STATUS_NOTFOUND;
+			error(logopt, MODPREFIX "failed to get query dn");
+			goto out;
 		}
 	}
 
-	return ldap;
+out:
+	return ret;
 }
 
 static unsigned long get_amd_timestamp(struct lookup_context *ctxt)
 {
+	struct ldap_conn conn;
 	LDAP *ldap;
 	LDAPMessage *result = NULL, *e;
 	char *query;
@@ -687,9 +710,11 @@ static unsigned long get_amd_timestamp(struct lookup_context *ctxt)
 	unsigned long timestamp = 0;
 	int rv, l, ql;
 
-	ldap = do_connect(LOGOPT_ANY, ctxt->server, ctxt);
-	if (!ldap)
+	memset(&conn, 0, sizeof(struct ldap_conn));
+	rv = do_connect(LOGOPT_ANY, &conn, ctxt->server, ctxt);
+	if (rv != NSS_STATUS_SUCCESS)
 		return 0;
+	ldap = conn.ldap;
 
 	map = amd_timestamp.map_attr;
 	class = amd_timestamp.entry_class;
@@ -726,7 +751,7 @@ static unsigned long get_amd_timestamp(struct lookup_context *ctxt)
 	rv = ldap_search_s(ldap, ctxt->base, scope, query, attrs, 0, &result);
 	if ((rv != LDAP_SUCCESS) || !result) {
 		crit(LOGOPT_ANY, MODPREFIX "timestamp query failed %s", query);
-		unbind_ldap_connection(LOGOPT_ANY, ldap, ctxt);
+		unbind_ldap_connection(LOGOPT_ANY, &conn, ctxt);
 		if (result)
 			ldap_msgfree(result);
 		free(query);
@@ -738,7 +763,7 @@ static unsigned long get_amd_timestamp(struct lookup_context *ctxt)
 		debug(LOGOPT_ANY,
 		     MODPREFIX "got answer, but no entry for timestamp");
 		ldap_msgfree(result);
-		unbind_ldap_connection(LOGOPT_ANY, ldap, ctxt);
+		unbind_ldap_connection(LOGOPT_ANY, &conn, ctxt);
 		free(query);
 		return CHE_MISSING;
 	}
@@ -789,61 +814,67 @@ next:
 	}
 
 	ldap_msgfree(result);
-	unbind_ldap_connection(LOGOPT_ANY, ldap, ctxt);
+	unbind_ldap_connection(LOGOPT_ANY, &conn, ctxt);
 	free(query);
 
 	return timestamp;
 }
 
-static LDAP *connect_to_server(unsigned logopt, const char *uri, struct lookup_context *ctxt)
+static int connect_to_server(unsigned logopt, struct ldap_conn *conn,
+			     const char *uri, struct lookup_context *ctxt)
 {
-	LDAP *ldap;
+	int ret;
 
-	ldap = do_connect(logopt, uri, ctxt);
-	if (!ldap) {
+	ret = do_connect(logopt, conn, uri, ctxt);
+	if (ret != NSS_STATUS_SUCCESS) {
 		warn(logopt,
 		     MODPREFIX "couldn't connect to server %s",
 		     uri ? uri : "default");
-		return NULL;
 	}
 
-	return ldap;
+	return ret;
 }
 
-static LDAP *find_dc_server(unsigned logopt, const char *uri, struct lookup_context *ctxt)
+static int find_dc_server(unsigned logopt, struct ldap_conn *conn,
+			  const char *uri, struct lookup_context *ctxt)
 {
 	char *str, *tok, *ptr = NULL;
-	LDAP *ldap = NULL;
+	int ret = NSS_STATUS_UNAVAIL;
 
 	str = strdup(uri);
 	if (!str)
-		return NULL;
+		return ret;
 
 	tok = strtok_r(str, " ", &ptr);
 	while (tok) {
 		const char *this = (const char *) tok;
+		int rv;
+
 		debug(logopt, "trying server uri %s", this);
-		ldap = connect_to_server(logopt, this, ctxt);
-		if (ldap) {
+		rv = connect_to_server(logopt, conn, this, ctxt);
+		if (rv == NSS_STATUS_SUCCESS) {
 			info(logopt, "connected to uri %s", this);
 			free(str);
-			return ldap;
+			return rv;
 		}
+		if (rv == NSS_STATUS_NOTFOUND)
+			ret = NSS_STATUS_NOTFOUND;
 		tok = strtok_r(NULL, " ", &ptr);
 	}
 
 	free(str);
 
-	return NULL;
+	return ret;
 }
 
-static LDAP *find_server(unsigned logopt, struct lookup_context *ctxt)
+static int find_server(unsigned logopt,
+		       struct ldap_conn *conn, struct lookup_context *ctxt)
 {
-	LDAP *ldap = NULL;
-	struct ldap_uri *this;
+	struct ldap_uri *this = NULL;
 	struct list_head *p, *first;
 	struct dclist *dclist;
 	char *uri = NULL;
+	int ret = NSS_STATUS_UNAVAIL;
 
 	uris_mutex_lock(ctxt);
 	dclist = ctxt->dclist;
@@ -857,6 +888,8 @@ static LDAP *find_server(unsigned logopt, struct lookup_context *ctxt)
 	/* Try each uri, save point in server list upon success */
 	p = first->next;
 	while(p != first) {
+		int rv;
+
 		/* Skip list head */
 		if (p == ctxt->uris) {
 			p = p->next;
@@ -866,12 +899,15 @@ static LDAP *find_server(unsigned logopt, struct lookup_context *ctxt)
 		if (!strstr(this->uri, ":///")) {
 			uri = strdup(this->uri);
 			debug(logopt, "trying server uri %s", uri);
-			ldap = connect_to_server(logopt, uri, ctxt);
-			if (ldap) {
+			rv = connect_to_server(logopt, conn, uri, ctxt);
+			if (rv == NSS_STATUS_SUCCESS) {
+				ret = NSS_STATUS_SUCCESS;
 				info(logopt, "connected to uri %s", uri);
 				free(uri);
 				break;
 			}
+			if (rv == NSS_STATUS_NOTFOUND)
+				ret = NSS_STATUS_NOTFOUND;
 		} else {
 			if (dclist)
 				uri = strdup(dclist->uri);
@@ -885,11 +921,14 @@ static LDAP *find_server(unsigned logopt, struct lookup_context *ctxt)
 				dclist = tmp;
 				uri = strdup(dclist->uri);
 			}
-			ldap = find_dc_server(logopt, uri, ctxt);
-			if (ldap) {
+			rv = find_dc_server(logopt, conn, uri, ctxt);
+			if (rv == NSS_STATUS_SUCCESS) {
+				ret = NSS_STATUS_SUCCESS;
 				free(uri);
 				break;
 			}
+			if (rv == NSS_STATUS_NOTFOUND)
+				ret = NSS_STATUS_NOTFOUND;
 		}
 		free(uri);
 		uri = NULL;
@@ -901,7 +940,7 @@ static LDAP *find_server(unsigned logopt, struct lookup_context *ctxt)
 	}
 
 	uris_mutex_lock(ctxt);
-	if (ldap)
+	if (conn->ldap)
 		ctxt->uri = this;
 	if (dclist) {
 		if (!ctxt->dclist)
@@ -915,36 +954,43 @@ static LDAP *find_server(unsigned logopt, struct lookup_context *ctxt)
 	}
 	uris_mutex_unlock(ctxt);
 
-	return ldap;
+	return ret;
 }
 
-static LDAP *do_reconnect(unsigned logopt, struct lookup_context *ctxt)
+static int do_reconnect(unsigned logopt,
+			struct ldap_conn *conn, struct lookup_context *ctxt)
 {
-	LDAP *ldap = NULL;
+	int ret = NSS_STATUS_UNAVAIL;
+	int dcrv = NSS_STATUS_SUCCESS;
+	int rv = NSS_STATUS_SUCCESS;
 
+	ldapinit_mutex_lock();
 	if (ctxt->server || !ctxt->uris) {
-		ldap = do_connect(logopt, ctxt->server, ctxt);
+		ret = do_connect(logopt, conn, ctxt->server, ctxt);
 #ifdef WITH_SASL
 		/* Dispose of the sasl authentication connection and try again. */
-		if (!ldap && ctxt->auth_required & LDAP_NEED_AUTH) {
-			ldapinit_mutex_lock();
-			autofs_sasl_dispose(ctxt);
-			ldapinit_mutex_unlock();
-			ldap = connect_to_server(logopt, ctxt->server, ctxt);
+		if (ctxt->auth_required & LDAP_NEED_AUTH &&
+		    ret != NSS_STATUS_SUCCESS && ret != NSS_STATUS_NOTFOUND) {
+			autofs_sasl_dispose(conn, ctxt);
+			ret = connect_to_server(logopt, conn,
+						ctxt->server, ctxt);
 		}
 #endif
-		return ldap;
+		ldapinit_mutex_unlock();
+		return ret;
 	}
 
 	if (ctxt->dclist) {
-		ldap = find_dc_server(logopt, ctxt->dclist->uri, ctxt);
-		if (ldap)
-			return ldap;
+		dcrv = find_dc_server(logopt, conn, ctxt->dclist->uri, ctxt);
+		if (dcrv == NSS_STATUS_SUCCESS) {
+			ldapinit_mutex_unlock();
+			return dcrv;
+		}
 	}
 
 	uris_mutex_lock(ctxt);
 	if (ctxt->dclist) {
-		if (!ldap || ctxt->dclist->expire < time(NULL)) {
+		if (!conn->ldap || ctxt->dclist->expire < monotonic_time(NULL)) {
 			free_dclist(ctxt->dclist);
 			ctxt->dclist = NULL;
 		}
@@ -958,37 +1004,42 @@ static LDAP *do_reconnect(unsigned logopt, struct lookup_context *ctxt)
 	if (!ctxt->uri)
 		goto find_server;
 
-	ldap = do_connect(logopt, ctxt->uri->uri, ctxt);
+	rv = do_connect(logopt, conn, ctxt->uri->uri, ctxt);
 #ifdef WITH_SASL
 	/*
 	 * Dispose of the sasl authentication connection and try the
 	 * current server again before trying other servers in the list.
 	 */
-	if (!ldap && ctxt->auth_required & LDAP_NEED_AUTH) {
-		ldapinit_mutex_lock();
-		autofs_sasl_dispose(ctxt);
-		ldapinit_mutex_unlock();
-		ldap = connect_to_server(logopt, ctxt->uri->uri, ctxt);
+	if (ctxt->auth_required & LDAP_NEED_AUTH &&
+	    rv != NSS_STATUS_SUCCESS && rv != NSS_STATUS_NOTFOUND) {
+		autofs_sasl_dispose(conn, ctxt);
+		rv = connect_to_server(logopt, conn, ctxt->uri->uri, ctxt);
 	}
 #endif
-	if (ldap)
-		return ldap;
+	if (rv == NSS_STATUS_SUCCESS) {
+		ldapinit_mutex_unlock();
+		return rv;
+	}
 
 	/* Failed to connect, try to find a new server */
 
 find_server:
 #ifdef WITH_SASL
-	ldapinit_mutex_lock();
-	autofs_sasl_dispose(ctxt);
-	ldapinit_mutex_unlock();
+	autofs_sasl_dispose(conn, ctxt);
 #endif
 
 	/* Current server failed, try the rest or dc connection */
-	ldap = find_server(logopt, ctxt);
-	if (!ldap)
+	ret = find_server(logopt, conn, ctxt);
+	if (ret != NSS_STATUS_SUCCESS) {
+		if (ret == NSS_STATUS_NOTFOUND ||
+		    dcrv == NSS_STATUS_NOTFOUND ||
+		    rv == NSS_STATUS_NOTFOUND)
+			ret = NSS_STATUS_NOTFOUND;
 		error(logopt, MODPREFIX "failed to find available server");
+	}
+	ldapinit_mutex_unlock();
 
-	return ldap;
+	return ret;
 }
 
 int get_property(unsigned logopt, xmlNodePtr node, const char *prop, char **value)
@@ -1626,38 +1677,23 @@ static void validate_uris(struct list_head *list)
 	return;			
 }
 
-/*
- * This initializes a context (persistent non-global data) for queries to
- * this module.  Return zero if we succeed.
- */
-int lookup_init(const char *mapfmt, int argc, const char *const *argv, void **context)
+static int do_init(const char *mapfmt,
+		   int argc, const char *const *argv,
+		   struct lookup_context *ctxt, unsigned int reinit)
 {
 	unsigned int is_amd_format;
-	struct lookup_context *ctxt;
-	char buf[MAX_ERR_BUF];
 	int ret;
 
-	*context = NULL;
-
-	/* If we can't build a context, bail. */
-	ctxt = malloc(sizeof(struct lookup_context));
-	if (!ctxt) {
-		char *estr = strerror_r(errno, buf, sizeof(buf));
-		logerr(MODPREFIX "malloc: %s", estr);
-		return 1;
-	}
-	memset(ctxt, 0, sizeof(struct lookup_context));
-
 	ret = pthread_mutex_init(&ctxt->uris_mutex, NULL);
 	if (ret) {
 		error(LOGOPT_ANY, MODPREFIX "failed to init uris mutex");
-		free(ctxt);
 		return 1;
 	}
 
 	/* If a map type isn't explicitly given, parse it like sun entries. */
 	if (mapfmt == NULL)
 		mapfmt = MAPFMT_DEFAULT;
+
 	is_amd_format = 0;
 	if (!strcmp(mapfmt, "amd")) {
 		is_amd_format = 1;
@@ -1675,7 +1711,6 @@ int lookup_init(const char *mapfmt, int argc, const char *const *argv, void **co
 		 */
 		if (!parse_server_string(LOGOPT_NONE, argv[0], ctxt)) {
 			error(LOGOPT_ANY, MODPREFIX "cannot parse server string");
-			free_context(ctxt);
 			return 1;
 		}
 
@@ -1700,7 +1735,6 @@ int lookup_init(const char *mapfmt, int argc, const char *const *argv, void **co
 		char *tmp = conf_amd_get_ldap_base();
 		if (!tmp) {
 			error(LOGOPT_ANY, MODPREFIX "failed to get base dn");
-			free_context(ctxt);
 			return 1;
 		}
 		ctxt->base = tmp;
@@ -1709,7 +1743,6 @@ int lookup_init(const char *mapfmt, int argc, const char *const *argv, void **co
 		if (!tmp) {
 			error(LOGOPT_ANY,
 			      MODPREFIX "failed to get ldap_hostports");
-			free_context(ctxt);
 			return 1;
 		}
 
@@ -1719,21 +1752,19 @@ int lookup_init(const char *mapfmt, int argc, const char *const *argv, void **co
 		 */
 		if (!parse_server_string(LOGOPT_NONE, tmp, ctxt)) {
 			error(LOGOPT_ANY, MODPREFIX "cannot parse server string");
-			free_context(ctxt);
+			free(tmp);
 			return 1;
 		}
 		free(tmp);
 
 		if (!ctxt->server) {
 			error(LOGOPT_ANY, MODPREFIX "ldap_hostports not valid");
-			free_context(ctxt);
 			return 1;
 		}
 
 		tmp = strdup(argv[0]);
 		if (!tmp) {
 			error(LOGOPT_ANY, MODPREFIX "failed to set mapname");
-			free_context(ctxt);
 			return 1;
 		}
 		ctxt->mapname = tmp;
@@ -1747,7 +1778,7 @@ int lookup_init(const char *mapfmt, int argc, const char *const *argv, void **co
 	 */
 	ret = parse_ldap_config(LOGOPT_NONE, ctxt);
 	if (ret) {
-		free_context(ctxt);
+		error(LOGOPT_ANY, MODPREFIX "failed to parse ldap config");
 		return 1;
 	}
 
@@ -1757,7 +1788,6 @@ int lookup_init(const char *mapfmt, int argc, const char *const *argv, void **co
 	if (!autofs_sasl_client_init(LOGOPT_NONE)) {
 		error(LOGOPT_ANY, "failed to init sasl client");
 		ldapinit_mutex_unlock();
-		free_context(ctxt);
 		return 1;
 	}
 	ldapinit_mutex_unlock();
@@ -1766,24 +1796,95 @@ int lookup_init(const char *mapfmt, int argc, const char *const *argv, void **co
 	if (is_amd_format)
 		ctxt->timestamp = get_amd_timestamp(ctxt);
 
-	/* Open the parser, if we can. */
-	ctxt->parse = open_parse(mapfmt, MODPREFIX, argc - 1, argv + 1);
-	if (!ctxt->parse) {
-		free_context(ctxt);
-		logerr(MODPREFIX "failed to open parse context");
+	if (reinit) {
+		ret = reinit_parse(ctxt->parse,
+				   mapfmt, MODPREFIX, argc - 1, argv + 1);
+		if (ret)
+			logmsg(MODPREFIX "failed to reinit parse context");
+	} else {
+		/* Open the parser, if we can. */
+		ctxt->parse = open_parse(mapfmt, MODPREFIX, argc - 1, argv + 1);
+		if (!ctxt->parse) {
+			logerr(MODPREFIX "failed to open parse context");
+			ret = 1;
+		}
+	}
+
+	return ret;
+}
+
+/*
+ * This initializes a context (persistent non-global data) for queries to
+ * this module.  Return zero if we succeed.
+ */
+int lookup_init(const char *mapfmt,
+		int argc, const char *const *argv, void **context)
+{
+	struct lookup_context *ctxt;
+	char buf[MAX_ERR_BUF];
+	int ret;
+
+	*context = NULL;
+
+	/* If we can't build a context, bail. */
+	ctxt = malloc(sizeof(struct lookup_context));
+	if (!ctxt) {
+		char *estr = strerror_r(errno, buf, sizeof(buf));
+		logerr(MODPREFIX "malloc: %s", estr);
 		return 1;
 	}
+	memset(ctxt, 0, sizeof(struct lookup_context));
+
+	ret = do_init(mapfmt, argc, argv, ctxt, 0);
+	if (ret) {
+		free_context(ctxt);
+		return 1;
+	}
+
 	*context = ctxt;
 
 	return 0;
 }
 
+int lookup_reinit(const char *mapfmt,
+		  int argc, const char *const *argv, void **context)
+{
+	struct lookup_context *ctxt = (struct lookup_context *) *context;
+	struct lookup_context *new;
+	char buf[MAX_ERR_BUF];
+	int ret;
+
+	/* If we can't build a context, bail. */
+	new = malloc(sizeof(struct lookup_context));
+	if (!new) {
+		char *estr = strerror_r(errno, buf, sizeof(buf));
+		logerr(MODPREFIX "malloc: %s", estr);
+		return 1;
+	}
+	memset(new, 0, sizeof(struct lookup_context));
+
+	new->parse = ctxt->parse;
+	ret = do_init(mapfmt, argc, argv, new, 1);
+	if (ret) {
+		free_context(new);
+		return 1;
+	}
+
+	*context = new;
+
+	free_context(ctxt);
+
+	return 0;
+}
+
 int lookup_read_master(struct master *master, time_t age, void *context)
 {
 	struct lookup_context *ctxt = (struct lookup_context *) context;
 	unsigned int timeout = master->default_timeout;
 	unsigned int logging = master->default_logging;
 	unsigned int logopt = master->logopt;
+	struct ldap_conn conn;
+	LDAP *ldap;
 	int rv, l, count;
 	char buf[MAX_ERR_BUF];
 	char parse_buf[PARSE_MAX_BUF];
@@ -1794,12 +1895,13 @@ int lookup_read_master(struct master *master, time_t age, void *context)
 	char **values = NULL;
 	char *attrs[3];
 	int scope = LDAP_SCOPE_SUBTREE;
-	LDAP *ldap;
 
 	/* Initialize the LDAP context. */
-	ldap = do_reconnect(logopt, ctxt);
-	if (!ldap)
-		return NSS_STATUS_UNAVAIL;
+	memset(&conn, 0, sizeof(struct ldap_conn));
+	rv = do_reconnect(logopt, &conn, ctxt);
+	if (rv)
+		return rv;
+	ldap = conn.ldap;
 
 	class = ctxt->schema->entry_class;
 	entry = ctxt->schema->entry_attr;
@@ -1833,7 +1935,7 @@ int lookup_read_master(struct master *master, time_t age, void *context)
 	if ((rv != LDAP_SUCCESS) || !result) {
 		error(logopt, MODPREFIX "query failed for %s: %s",
 		      query, ldap_err2string(rv));
-		unbind_ldap_connection(logging, ldap, ctxt);
+		unbind_ldap_connection(logging, &conn, ctxt);
 		if (result)
 			ldap_msgfree(result);
 		free(query);
@@ -1846,7 +1948,7 @@ int lookup_read_master(struct master *master, time_t age, void *context)
 		      MODPREFIX "query succeeded, no matches for %s",
 		      query);
 		ldap_msgfree(result);
-		unbind_ldap_connection(logging, ldap, ctxt);
+		unbind_ldap_connection(logging, &conn, ctxt);
 		free(query);
 		return NSS_STATUS_NOTFOUND;
 	} else
@@ -1967,7 +2069,7 @@ next:
 
 	/* Clean up. */
 	ldap_msgfree(result);
-	unbind_ldap_connection(logopt, ldap, ctxt);
+	unbind_ldap_connection(logopt, &conn, ctxt);
 	free(query);
 
 	return NSS_STATUS_SUCCESS;
@@ -2687,6 +2789,7 @@ static int read_one_map(struct autofs_point *ap,
 			struct lookup_context *ctxt,
 			time_t age, int *result_ldap)
 {
+	struct ldap_conn conn;
 	struct ldap_search_params sp;
 	char buf[MAX_ERR_BUF];
 	char *class, *info, *entry;
@@ -2707,9 +2810,11 @@ static int read_one_map(struct autofs_point *ap,
 	sp.age = age;
 
 	/* Initialize the LDAP context. */
-	sp.ldap = do_reconnect(ap->logopt, ctxt);
-	if (!sp.ldap)
-		return NSS_STATUS_UNAVAIL;
+	memset(&conn, 0, sizeof(struct ldap_conn));
+	rv = do_reconnect(ap->logopt, &conn, ctxt);
+	if (rv)
+		return rv;
+	sp.ldap = conn.ldap;
 
 	class = ctxt->schema->entry_class;
 	entry = ctxt->schema->entry_attr;
@@ -2768,7 +2873,7 @@ static int read_one_map(struct autofs_point *ap,
 			if (sp.pageSize < 5) {
 				debug(ap->logopt, MODPREFIX
 				      "result size too small");
-				unbind_ldap_connection(ap->logopt, sp.ldap, ctxt);
+				unbind_ldap_connection(ap->logopt, &conn, ctxt);
 				*result_ldap = rv;
 				free(sp.query);
 				return NSS_STATUS_UNAVAIL;
@@ -2777,7 +2882,7 @@ static int read_one_map(struct autofs_point *ap,
 		}
 
 		if (rv != LDAP_SUCCESS || !sp.result) {
-			unbind_ldap_connection(ap->logopt, sp.ldap, ctxt);
+			unbind_ldap_connection(ap->logopt, &conn, ctxt);
 			*result_ldap = rv;
 			if (sp.result)
 				ldap_msgfree(sp.result);
@@ -2793,7 +2898,7 @@ static int read_one_map(struct autofs_point *ap,
 			rv = do_get_entries(&sp, source, ctxt);
 		if (rv != LDAP_SUCCESS) {
 			ldap_msgfree(sp.result);
-			unbind_ldap_connection(ap->logopt, sp.ldap, ctxt);
+			unbind_ldap_connection(ap->logopt, &conn, ctxt);
 			*result_ldap = rv;
 			if (sp.cookie)
 				ber_bvfree(sp.cookie);
@@ -2806,7 +2911,7 @@ static int read_one_map(struct autofs_point *ap,
 
 	debug(ap->logopt, MODPREFIX "done updating map");
 
-	unbind_ldap_connection(ap->logopt, sp.ldap, ctxt);
+	unbind_ldap_connection(ap->logopt, &conn, ctxt);
 
 	source->age = age;
 	if (sp.cookie)
@@ -2849,9 +2954,11 @@ static int lookup_one(struct autofs_point *ap, struct map_source *source,
 		char *qKey, int qKey_len, struct lookup_context *ctxt)
 {
 	struct mapent_cache *mc;
+	struct ldap_conn conn;
+	LDAP *ldap;
 	int rv, i, l, ql, count;
 	char buf[MAX_ERR_BUF];
-	time_t age = time(NULL);
+	time_t age = monotonic_time(NULL);
 	char *query;
 	LDAPMessage *result = NULL, *e;
 	char *class, *info, *entry;
@@ -2861,7 +2968,6 @@ static int lookup_one(struct autofs_point *ap, struct map_source *source,
 	struct berval **bvValues;
 	char *attrs[3];
 	int scope = LDAP_SCOPE_SUBTREE;
-	LDAP *ldap;
 	struct mapent *we;
 	unsigned int wild = 0;
 	int ret = CHE_MISSING;
@@ -2874,9 +2980,13 @@ static int lookup_one(struct autofs_point *ap, struct map_source *source,
 	}
 
 	/* Initialize the LDAP context. */
-	ldap = do_reconnect(ap->logopt, ctxt);
-	if (!ldap)
+	memset(&conn, 0, sizeof(struct ldap_conn));
+	rv = do_reconnect(ap->logopt, &conn, ctxt);
+	if (rv == NSS_STATUS_UNAVAIL)
 		return CHE_UNAVAIL;
+	if (rv == NSS_STATUS_NOTFOUND)
+		return ret;
+	ldap = conn.ldap;
 
 	class = ctxt->schema->entry_class;
 	entry = ctxt->schema->entry_attr;
@@ -2964,7 +3074,7 @@ static int lookup_one(struct autofs_point *ap, struct map_source *source,
 
 	if ((rv != LDAP_SUCCESS) || !result) {
 		crit(ap->logopt, MODPREFIX "query failed for %s", query);
-		unbind_ldap_connection(ap->logopt, ldap, ctxt);
+		unbind_ldap_connection(ap->logopt, &conn, ctxt);
 		if (result)
 			ldap_msgfree(result);
 		free(query);
@@ -2979,7 +3089,7 @@ static int lookup_one(struct autofs_point *ap, struct map_source *source,
 		debug(ap->logopt,
 		     MODPREFIX "got answer, but no entry for %s", query);
 		ldap_msgfree(result);
-		unbind_ldap_connection(ap->logopt, ldap, ctxt);
+		unbind_ldap_connection(ap->logopt, &conn, ctxt);
 		free(query);
 		return CHE_MISSING;
 	}
@@ -3165,7 +3275,7 @@ next:
 	}
 
 	ldap_msgfree(result);
-	unbind_ldap_connection(ap->logopt, ldap, ctxt);
+	unbind_ldap_connection(ap->logopt, &conn, ctxt);
 
 	/* Failed to find wild entry, update cache if needed */
 	cache_writelock(mc);
@@ -3205,6 +3315,7 @@ static int lookup_one_amd(struct autofs_point *ap,
 			  struct lookup_context *ctxt)
 {
 	struct mapent_cache *mc = source->mc;
+	struct ldap_conn conn;
 	LDAP *ldap;
 	LDAPMessage *result = NULL, *e;
 	char *query;
@@ -3214,7 +3325,7 @@ static int lookup_one_amd(struct autofs_point *ap,
 	struct berval **bvKey;
 	struct berval **bvValues;
 	char buf[MAX_ERR_BUF];
-	time_t age = time(NULL);
+	time_t age = monotonic_time(NULL);
 	int rv, l, ql, count;
 	int ret = CHE_MISSING;
 
@@ -3224,9 +3335,13 @@ static int lookup_one_amd(struct autofs_point *ap,
 	}
 
 	/* Initialize the LDAP context. */
-	ldap = do_reconnect(ap->logopt, ctxt);
-	if (!ldap)
+	memset(&conn, 0, sizeof(struct ldap_conn));
+	rv = do_reconnect(ap->logopt, &conn, ctxt);
+	if (rv == NSS_STATUS_UNAVAIL)
 		return CHE_UNAVAIL;
+	if (rv == NSS_STATUS_NOTFOUND)
+		return ret;
+	ldap = conn.ldap;
 
 	map = ctxt->schema->map_attr;
 	class = ctxt->schema->entry_class;
@@ -3268,7 +3383,7 @@ static int lookup_one_amd(struct autofs_point *ap,
 	rv = ldap_search_s(ldap, ctxt->base, scope, query, attrs, 0, &result);
 	if ((rv != LDAP_SUCCESS) || !result) {
 		crit(ap->logopt, MODPREFIX "query failed for %s", query);
-		unbind_ldap_connection(ap->logopt, ldap, ctxt);
+		unbind_ldap_connection(ap->logopt, &conn, ctxt);
 		if (result)
 			ldap_msgfree(result);
 		free(query);
@@ -3283,7 +3398,7 @@ static int lookup_one_amd(struct autofs_point *ap,
 		debug(ap->logopt,
 		     MODPREFIX "got answer, but no entry for %s", query);
 		ldap_msgfree(result);
-		unbind_ldap_connection(ap->logopt, ldap, ctxt);
+		unbind_ldap_connection(ap->logopt, &conn, ctxt);
 		free(query);
 		return CHE_MISSING;
 	}
@@ -3345,7 +3460,7 @@ next:
 	}
 
 	ldap_msgfree(result);
-	unbind_ldap_connection(ap->logopt, ldap, ctxt);
+	unbind_ldap_connection(ap->logopt, &conn, ctxt);
 	free(query);
 
 	return ret;
@@ -3417,7 +3532,7 @@ static int check_map_indirect(struct autofs_point *ap,
 	unsigned int is_amd_format = source->flags & MAP_FLAG_FORMAT_AMD;
 	struct mapent_cache *mc;
 	struct mapent *me;
-	time_t now = time(NULL);
+	time_t now = monotonic_time(NULL);
 	time_t t_last_read;
 	int ret, cur_state;
 	int status;
@@ -3555,7 +3670,7 @@ int lookup_mount(struct autofs_point *ap, const char *name, int name_len, void *
 	/* Check if we recorded a mount fail for this key anywhere */
 	me = lookup_source_mapent(ap, key, LKP_DISTINCT);
 	if (me) {
-		if (me->status >= time(NULL)) {
+		if (me->status >= monotonic_time(NULL)) {
 			cache_unlock(me->mc);
 			return NSS_STATUS_NOTFOUND;
 		} else {
@@ -3702,7 +3817,7 @@ int lookup_done(void *context)
 	int rv = close_parse(ctxt->parse);
 #ifdef WITH_SASL
 	ldapinit_mutex_lock();
-	autofs_sasl_dispose(ctxt);
+	autofs_sasl_dispose(NULL, ctxt);
 	autofs_sasl_done();
 	ldapinit_mutex_unlock();
 #endif
