@@ -49,7 +49,63 @@ struct parse_context {
 
 int lookup_version = AUTOFS_LOOKUP_VERSION;	/* Required by protocol */
 
-int lookup_init(const char *mapfmt, int argc, const char *const *argv, void **context)
+static int do_init(const char *mapfmt,
+		   int argc, const char *const *argv,
+		   struct lookup_context *ctxt, unsigned int reinit)
+{
+	int ret = 0;
+
+	if (argc < 1) {
+		logmsg(MODPREFIX "No map name");
+		ret = 1;
+		goto out;
+	}
+	ctxt->mapname = argv[0];
+
+	if (ctxt->mapname[0] != '/') {
+		logmsg(MODPREFIX "program map %s is not an absolute pathname",
+		     ctxt->mapname);
+		ret = 1;
+		goto out;
+	}
+
+	if (access(ctxt->mapname, X_OK)) {
+		logmsg(MODPREFIX "program map %s missing or not executable",
+		     ctxt->mapname);
+		ret = 1;
+		goto out;
+	}
+
+	if (!mapfmt)
+		mapfmt = MAPFMT_DEFAULT;
+
+	ctxt->mapfmt = strdup(mapfmt);
+	if (!ctxt->mapfmt) {
+		logmsg(MODPREFIX "failed to allocate storage for map format");
+		ret = 1;
+		goto out;
+	}
+
+	if (reinit) {
+		ret = reinit_parse(ctxt->parse, mapfmt, MODPREFIX, argc - 1, argv + 1);
+		if (ret)
+			logmsg(MODPREFIX "failed to reinit parse context");
+	} else {
+		ctxt->parse = open_parse(mapfmt, MODPREFIX, argc - 1, argv + 1);
+		if (!ctxt->parse) {
+			logmsg(MODPREFIX "failed to open parse context");
+			ret = 1;
+		}
+	}
+out:
+	if (ret && ctxt->mapfmt)
+		free(ctxt->mapfmt);
+
+	return ret;
+}
+
+int lookup_init(const char *mapfmt,
+		int argc, const char *const *argv, void **context)
 {
 	struct lookup_context *ctxt;
 	char buf[MAX_ERR_BUF];
@@ -62,44 +118,49 @@ int lookup_init(const char *mapfmt, int argc, const char *const *argv, void **co
 		logerr(MODPREFIX "malloc: %s", estr);
 		return 1;
 	}
+	memset(ctxt, 0, sizeof(struct lookup_context));
 
-	if (argc < 1) {
-		logmsg(MODPREFIX "No map name");
-		free(ctxt);
-		return 1;
-	}
-	ctxt->mapname = argv[0];
-
-	if (ctxt->mapname[0] != '/') {
-		logmsg(MODPREFIX "program map %s is not an absolute pathname",
-		     ctxt->mapname);
+	if (do_init(mapfmt, argc, argv, ctxt, 0)) {
 		free(ctxt);
 		return 1;
 	}
 
-	if (access(ctxt->mapname, X_OK)) {
-		logmsg(MODPREFIX "program map %s missing or not executable",
-		     ctxt->mapname);
-		free(ctxt);
-		return 1;
-	}
-
-	if (!mapfmt)
-		mapfmt = MAPFMT_DEFAULT;
-
-	ctxt->mapfmt = strdup(mapfmt);
-
-	ctxt->parse = open_parse(mapfmt, MODPREFIX, argc - 1, argv + 1);
-	if (!ctxt->parse) {
-		logmsg(MODPREFIX "failed to open parse context");
-		free(ctxt);
-		return 1;
-	}
 	*context = ctxt;
 
 	return 0;
 }
 
+int lookup_reinit(const char *mapfmt,
+		  int argc, const char *const *argv, void **context)
+{
+	struct lookup_context *ctxt = (struct lookup_context *) *context;
+	struct lookup_context *new;
+	char buf[MAX_ERR_BUF];
+	int ret;
+
+	new = malloc(sizeof(struct lookup_context));
+	if (!new) {
+		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
+		logerr(MODPREFIX "malloc: %s", estr);
+		return 1;
+	}
+	memset(new, 0, sizeof(struct lookup_context));
+
+	new->parse = ctxt->parse;
+	ret = do_init(mapfmt, argc, argv, new, 1);
+	if (ret) {
+		free(new);
+		return 1;
+	}
+
+	*context = new;
+
+	free(ctxt->mapfmt);
+	free(ctxt);
+
+	return 0;
+}
+
 int lookup_read_master(struct master *master, time_t age, void *context)
 {
         return NSS_STATUS_UNKNOWN;
@@ -129,6 +190,7 @@ static char *lookup_one(struct autofs_point *ap,
 	int distance;
 	int alloci = 1;
 	int status;
+	char *prefix;
 
 	mapent = (char *) malloc(MAPENT_MAX_LEN + 1);
 	if (!mapent) {
@@ -137,6 +199,16 @@ static char *lookup_one(struct autofs_point *ap,
 		return NULL;
 	}
 
+	/*
+	 * By default use a prefix with standard environment
+	 * variables to prevent system subversion by interpreted
+	 * languages.
+	 */
+	if (defaults_force_std_prog_map_env())
+		prefix = NULL;
+	else
+		prefix = "AUTOFS_";
+
 	/*
 	 * We don't use popen because we don't want to run /bin/sh plus we
 	 * want to send stderr to the syslog, and we don't use spawnl()
@@ -174,14 +246,15 @@ static char *lookup_one(struct autofs_point *ap,
 			warn(ap->logopt,
 			     MODPREFIX "failed to set PWD to %s for map %s",
 			     ap->path, ctxt->mapname);
+
 		/*
 		 * MAPFMT_DEFAULT must be "sun" for ->parse_init() to have setup
 		 * the macro table.
 		 */
-		if (ctxt->mapfmt && strcmp(ctxt->mapfmt, "MAPFMT_DEFAULT")) {
+		if (ctxt->mapfmt && !strcmp(ctxt->mapfmt, MAPFMT_DEFAULT)) {
 			struct parse_context *pctxt = (struct parse_context *) ctxt->parse->context;
 			/* Add standard environment as seen by sun map parser */
-			pctxt->subst = addstdenv(pctxt->subst);
+			pctxt->subst = addstdenv(pctxt->subst, prefix);
 			macro_setenv(pctxt->subst);
 		}
 		execl(ctxt->mapname, ctxt->mapname, name, NULL);
@@ -357,7 +430,7 @@ static int lookup_amd_defaults(struct autofs_point *ap,
 		while (isblank(*start))
 			start++;
 		cache_writelock(mc);
-		ret = cache_update(mc, source, "/defaults", start, time(NULL));
+		ret = cache_update(mc, source, "/defaults", start, monotonic_time(NULL));
 		cache_unlock(mc);
 		if (ret == CHE_FAIL) {
 			free(ment);
@@ -426,7 +499,7 @@ static int match_key(struct autofs_point *ap,
 				start++;
 		}
 		cache_writelock(mc);
-		ret = cache_update(mc, source, lkp_key, start, time(NULL));
+		ret = cache_update(mc, source, lkp_key, start, monotonic_time(NULL));
 		cache_unlock(mc);
 		if (ret == CHE_FAIL) {
 			free(ment);
@@ -479,7 +552,7 @@ static int match_key(struct autofs_point *ap,
 			while (isblank(*start))
 				start++;
 			cache_writelock(mc);
-			ret = cache_update(mc, source, match, start, time(NULL));
+			ret = cache_update(mc, source, match, start, monotonic_time(NULL));
 			cache_unlock(mc);
 			if (ret == CHE_FAIL) {
 				free(match);
@@ -525,7 +598,7 @@ int lookup_mount(struct autofs_point *ap, const char *name, int name_len, void *
 	/* Check if we recorded a mount fail for this key anywhere */
 	me = lookup_source_mapent(ap, name, LKP_DISTINCT);
 	if (me) {
-		if (me->status >= time(NULL)) {
+		if (me->status >= monotonic_time(NULL)) {
 			cache_unlock(me->mc);
 			return NSS_STATUS_NOTFOUND;
 		} else {
@@ -574,7 +647,7 @@ int lookup_mount(struct autofs_point *ap, const char *name, int name_len, void *
 		 * proceed with the program map lookup.
 		 */
 		if (strchr(name, '/') ||
-		    me->age + ap->negative_timeout > time(NULL)) {
+		    me->age + ap->negative_timeout > monotonic_time(NULL)) {
 			char *ent = NULL;
 
 			if (me->mapent) {
@@ -622,20 +695,11 @@ out_free:
 		free(mapent);
 
 	if (ret) {
-		time_t now = time(NULL);
-		int rv = CHE_OK;
-
+		/* Don't update negative cache when re-connecting */
+		if (ap->flags & MOUNT_FLAG_REMOUNT)
+			return NSS_STATUS_TRYAGAIN;
 		cache_writelock(mc);
-		me = cache_lookup_distinct(mc, name);
-		if (me)
-			rv = cache_push_mapent(me, NULL);
-		else
-			rv = cache_update(mc, source, name, NULL, now);
-		if (rv != CHE_FAIL) {
-			me = cache_lookup_distinct(mc, name);
-			if (me)
-				me->status = now + ap->negative_timeout;
-		}
+		cache_update_negative(mc, source, name, ap->negative_timeout);
 		cache_unlock(mc);
 		return NSS_STATUS_TRYAGAIN;
 	}
