@@ -56,38 +56,16 @@ struct lookup_context {
 
 int lookup_version = AUTOFS_LOOKUP_VERSION;	/* Required by protocol */
 
-int lookup_init(const char *mapfmt, int argc, const char *const *argv, void **context)
+static int open_sss_lib(struct lookup_context *ctxt)
 {
-	struct lookup_context *ctxt;
-	char buf[MAX_ERR_BUF];
 	char dlbuf[PATH_MAX];
 	char *estr;
 	void *dh;
 	size_t size;
 
-	*context = NULL;
-
-	ctxt = malloc(sizeof(struct lookup_context));
-	if (!ctxt) {
-		estr = strerror_r(errno, buf, MAX_ERR_BUF);
-		logerr(MODPREFIX "malloc: %s", estr);
-		return 1;
-	}
-
-	if (argc < 1) {
-		free(ctxt);
-		logerr(MODPREFIX "No map name");
-		return 1;
-	}
-	ctxt->mapname = argv[0];
-
-	if (!mapfmt)
-		mapfmt = MAPFMT_DEFAULT;
-
 	size = snprintf(dlbuf, sizeof(dlbuf),
 			"%s/%s.so", SSS_LIB_DIR, SSS_SO_NAME);
 	if (size >= sizeof(dlbuf)) {
-		free(ctxt);
 		logmsg(MODPREFIX "sss library path too long");
 		return 1;
 	}
@@ -95,7 +73,6 @@ int lookup_init(const char *mapfmt, int argc, const char *const *argv, void **co
 	dh = dlopen(dlbuf, RTLD_LAZY);
 	if (!dh) {
 		logerr(MODPREFIX "failed to open %s: %s", dlbuf, dlerror());
-		free(ctxt);
 		return 1;
 	}
 	ctxt->dlhandle = dh;
@@ -116,15 +93,6 @@ int lookup_init(const char *mapfmt, int argc, const char *const *argv, void **co
 	if (!ctxt->setautomntent)
 		goto lib_names_fail;
 
-	ctxt->parse = open_parse(mapfmt, MODPREFIX, argc - 1, argv + 1);
-	if (!ctxt->parse) {
-		logmsg(MODPREFIX "failed to open parse context");
-		dlclose(dh);
-		free(ctxt);
-		return 1;
-	}
-	*context = ctxt;
-
 	return 0;
 
 lib_names_fail:
@@ -133,10 +101,109 @@ lib_names_fail:
 	else
 		logerr(MODPREFIX "dlsym: %s", estr);
 	dlclose(dh);
-	free(ctxt);
+
 	return 1;
 }
 
+static int do_init(const char *mapfmt,
+		   int argc, const char *const *argv,
+		   struct lookup_context *ctxt, unsigned int reinit)
+{
+	int ret = 0;
+
+	if (argc < 1) {
+		logerr(MODPREFIX "No map name");
+		ret = 1;
+		goto out;
+	}
+	ctxt->mapname = argv[0];
+
+	if (!mapfmt)
+		mapfmt = MAPFMT_DEFAULT;
+
+	if (!reinit) {
+		ret = open_sss_lib(ctxt);
+		if (ret)
+			goto out;
+	}
+
+	if (reinit) {
+		ret = reinit_parse(ctxt->parse, mapfmt, MODPREFIX, argc - 1, argv + 1);
+		if (ret)
+			logmsg(MODPREFIX "failed to reinit parse context");
+	} else {
+		ctxt->parse = open_parse(mapfmt, MODPREFIX, argc - 1, argv + 1);
+		if (!ctxt->parse) {
+			logmsg(MODPREFIX "failed to open parse context");
+			dlclose(ctxt->dlhandle);
+			ret = 1;
+		}
+	}
+out:
+	return ret;
+}
+
+int lookup_init(const char *mapfmt,
+		int argc, const char *const *argv, void **context)
+{
+	struct lookup_context *ctxt;
+	char buf[MAX_ERR_BUF];
+	char *estr;
+
+	*context = NULL;
+
+	ctxt = malloc(sizeof(struct lookup_context));
+	if (!ctxt) {
+		estr = strerror_r(errno, buf, MAX_ERR_BUF);
+		logerr(MODPREFIX "malloc: %s", estr);
+		return 1;
+	}
+
+	if (do_init(mapfmt, argc, argv, ctxt, 0)) {
+		free(ctxt);
+		return 1;
+	}
+
+	*context = ctxt;
+
+	return 0;
+}
+
+int lookup_reinit(const char *mapfmt,
+		  int argc, const char *const *argv, void **context)
+{
+	struct lookup_context *ctxt = (struct lookup_context *) *context;
+	struct lookup_context *new;
+	char buf[MAX_ERR_BUF];
+	int ret;
+
+	new = malloc(sizeof(struct lookup_context));
+	if (!new) {
+		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
+		logerr(MODPREFIX "malloc: %s", estr);
+		return 1;
+	}
+
+	new->parse = ctxt->parse;
+	ret = do_init(mapfmt, argc, argv, new, 1);
+	if (ret) {
+		free(new);
+		return 1;
+	}
+
+	new->dlhandle = ctxt->dlhandle;
+	new->setautomntent = ctxt->setautomntent;
+	new->getautomntent_r = ctxt->getautomntent_r;
+	new->getautomntbyname_r = ctxt->getautomntbyname_r;
+	new->endautomntent = ctxt->endautomntent;
+
+	*context = new;
+
+	free(ctxt);
+
+	return 0;
+}
+
 static int setautomntent(unsigned int logopt,
 			 struct lookup_context *ctxt, const char *mapname,
 			 void **sss_ctxt)
@@ -148,9 +215,8 @@ static int setautomntent(unsigned int logopt,
 		error(logopt, MODPREFIX "setautomntent: %s", estr);
 		if (*sss_ctxt)
 			free(*sss_ctxt);
-		return 0;
 	}
-	return 1;
+	return ret;
 }
 
 static int endautomntent(unsigned int logopt,
@@ -161,9 +227,8 @@ static int endautomntent(unsigned int logopt,
 		char buf[MAX_ERR_BUF];
 		char *estr = strerror_r(ret, buf, MAX_ERR_BUF);
 		error(logopt, MODPREFIX "endautomntent: %s", estr);
-		return 0;
 	}
-	return 1;
+	return ret;
 }
 
 int lookup_read_master(struct master *master, time_t age, void *context)
@@ -180,8 +245,12 @@ int lookup_read_master(struct master *master, time_t age, void *context)
 	char *value = NULL;
 	int count, ret;
 
-	if (!setautomntent(logopt, ctxt, ctxt->mapname, &sss_ctxt))
+	ret = setautomntent(logopt, ctxt, ctxt->mapname, &sss_ctxt);
+	if (ret) {
+		if (ret == ENOENT)
+			return NSS_STATUS_NOTFOUND;
 		return NSS_STATUS_UNAVAIL;
+	}
 
 	count = 0;
 	while (1) {
@@ -280,8 +349,12 @@ int lookup_read_map(struct autofs_point *ap, time_t age, void *context)
 		return NSS_STATUS_SUCCESS;
 	}
 
-	if (!setautomntent(ap->logopt, ctxt, ctxt->mapname, &sss_ctxt))
+	ret = setautomntent(ap->logopt, ctxt, ctxt->mapname, &sss_ctxt);
+	if (ret) {
+		if (ret == ENOENT)
+			return NSS_STATUS_NOTFOUND;
 		return NSS_STATUS_UNAVAIL;
+	}
 
 	count = 0;
 	while (1) {
@@ -374,7 +447,7 @@ static int lookup_one(struct autofs_point *ap,
 	struct mapent_cache *mc;
 	struct mapent *we;
 	void *sss_ctxt = NULL;
-	time_t age = time(NULL);
+	time_t age = monotonic_time(NULL);
 	char buf[MAX_ERR_BUF];
 	char *value = NULL;
 	char *s_key;
@@ -386,8 +459,12 @@ static int lookup_one(struct autofs_point *ap,
 
 	mc = source->mc;
 
-	if (!setautomntent(ap->logopt, ctxt, ctxt->mapname, &sss_ctxt))
+	ret = setautomntent(ap->logopt, ctxt, ctxt->mapname, &sss_ctxt);
+	if (ret) {
+		if (ret == ENOENT)
+			return NSS_STATUS_NOTFOUND;
 		return NSS_STATUS_UNAVAIL;
+	}
 
 	ret = ctxt->getautomntbyname_r(qKey, &value, sss_ctxt);
 	if (ret && ret != ENOENT) {
@@ -491,7 +568,7 @@ static int check_map_indirect(struct autofs_point *ap,
 	struct map_source *source;
 	struct mapent_cache *mc;
 	struct mapent *me;
-	time_t now = time(NULL);
+	time_t now = monotonic_time(NULL);
 	time_t t_last_read;
 	int ret, cur_state;
 
@@ -585,7 +662,7 @@ int lookup_mount(struct autofs_point *ap, const char *name, int name_len, void *
 	/* Check if we recorded a mount fail for this key anywhere */
 	me = lookup_source_mapent(ap, key, LKP_DISTINCT);
 	if (me) {
-		if (me->status >= time(NULL)) {
+		if (me->status >= monotonic_time(NULL)) {
 			cache_unlock(me->mc);
 			return NSS_STATUS_NOTFOUND;
 		} else {
@@ -599,8 +676,10 @@ int lookup_mount(struct autofs_point *ap, const char *name, int name_len, void *
 				cache_writelock(smc);
 				sme = cache_lookup_distinct(smc, key);
 				/* Negative timeout expired for non-existent entry. */
-				if (sme && !sme->mapent)
-					cache_delete(smc, key);
+				if (sme && !sme->mapent) {
+					if (cache_pop_mapent(sme) == CHE_FAIL)
+						cache_delete(smc, key);
+				}
 				cache_unlock(smc);
 			}
 		}
@@ -635,7 +714,17 @@ int lookup_mount(struct autofs_point *ap, const char *name, int name_len, void *
 			return status;
 	}
 
-	cache_readlock(mc);
+	/*
+	 * We can't take the writelock for direct mounts. If we're
+	 * starting up or trying to re-connect to an existing direct
+	 * mount we could be iterating through the map entries with
+	 * the readlock held. But we don't need to update the cache
+	 * when we're starting up so just take the readlock in that
+	 */
+	if (ap->flags & MOUNT_FLAG_REMOUNT)
+		cache_writelock(mc);
+	else
+		cache_readlock(mc);
 	me = cache_lookup(mc, key);
 	/* Stale mapent => check for entry in alternate source or wildcard */
 	if (me && !me->mapent) {
@@ -645,9 +734,21 @@ int lookup_mount(struct autofs_point *ap, const char *name, int name_len, void *
 		if (!me)
 			me = cache_lookup_distinct(mc, "*");
 	}
-	if (me && me->mapent && (me->source == source || *me->key == '/')) {
-		strcpy(mapent_buf, me->mapent);
-		mapent = mapent_buf;
+	if (me && me->mapent) {
+		/*
+		 * If this is a lookup add wildcard match for later validation
+		 * checks and negative cache lookups.
+		 */
+		if (ap->type == LKP_INDIRECT && *me->key == '*' &&
+		   !(ap->flags & MOUNT_FLAG_REMOUNT)) {
+			ret = cache_update(mc, source, key, me->mapent, me->age);
+			if (!(ret & (CHE_OK | CHE_UPDATED)))
+				me = NULL;
+		}
+		if (me && (me->source == source || *me->key == '/')) {
+			strcpy(mapent_buf, me->mapent);
+			mapent = mapent_buf;
+		}
 	}
 	cache_unlock(mc);
 
@@ -661,18 +762,11 @@ int lookup_mount(struct autofs_point *ap, const char *name, int name_len, void *
 	ret = ctxt->parse->parse_mount(ap, key, key_len,
 				       mapent, ctxt->parse->context);
 	if (ret) {
-		time_t now = time(NULL);
-		int rv = CHE_OK;
-
-		/* Record the the mount fail in the cache */
+		/* Don't update negative cache when re-connecting */
+		if (ap->flags & MOUNT_FLAG_REMOUNT)
+			return NSS_STATUS_TRYAGAIN;
 		cache_writelock(mc);
-		me = cache_lookup_distinct(mc, key);
-		if (!me)
-			rv = cache_update(mc, source, key, NULL, now);
-		if (rv != CHE_FAIL) {
-			me = cache_lookup_distinct(mc, key);
-			me->status = now + ap->negative_timeout;
-		}
+		cache_update_negative(mc, source, key, ap->negative_timeout);
 		cache_unlock(mc);
 		return NSS_STATUS_TRYAGAIN;
 	}
