@@ -39,6 +39,11 @@ int mount_init(void **context)
 	return 0;
 }
 
+int mount_reinit(void **context)
+{
+	return 0;
+}
+
 int mount_mount(struct autofs_point *ap, const char *root, const char *name,
 		int name_len, const char *what, const char *fstype,
 		const char *c_options, void *context)
@@ -203,6 +208,30 @@ int mount_mount(struct autofs_point *ap, const char *root, const char *name,
 	}
 	if (info->map)
 		argv[0] = info->map;
+	/*
+	 * If the parent map format is amd and the format isn't
+	 * specified in the map entry set it from the parent map
+	 * source.
+	 */
+	if (!info->format && ap->entry->maps) {
+		struct map_source *s = ap->entry->maps;
+		/*
+		 * For amd maps, if the format and source type aren't
+		 * specified try and set them from the parent.
+		 */
+		if (s->flags & MAP_FLAG_FORMAT_AMD) {
+			info->format = strdup("amd");
+			if (!info->format)
+				warn(ap->logopt, MODPREFIX
+				     "failed to set amd map format");
+			if (!info->type && s->type) {
+				info->type = strdup(s->type);
+				if (!info->type)
+					warn(ap->logopt, MODPREFIX
+					     "failed to set amd map type");
+			}
+		}
+	}
 
 	if (options) {
 		p = options;
@@ -218,7 +247,7 @@ int mount_mount(struct autofs_point *ap, const char *root, const char *name,
 
 	source = master_add_map_source(entry,
 				       info->type, info->format,
-				       time(NULL), argc, argv);
+				       monotonic_time(NULL), argc, argv);
 	if (!source) {
 		error(ap->logopt,
 		      MODPREFIX "failed to add map source to entry");
@@ -231,6 +260,16 @@ int mount_mount(struct autofs_point *ap, const char *root, const char *name,
 
 	mounts_mutex_lock(ap);
 
+	if (source->flags & MAP_FLAG_FORMAT_AMD) {
+		struct amd_entry *am_entry = __master_find_amdmount(ap, entry->path);
+		if (am_entry && am_entry->pref) {
+			nap->pref = am_entry->pref;
+			am_entry->pref = NULL;
+		}
+		/* amd mounts don't support browse mode */
+		nap->flags &= ~MOUNT_FLAG_GHOST;
+	}
+
 	if (handle_mounts_startup_cond_init(&suc)) {
 		crit(ap->logopt, MODPREFIX
 		     "failed to init startup cond for mount %s", entry->path);
