@@ -15,8 +15,6 @@
 
 #include <stdio.h>
 #include <malloc.h>
-#include <fcntl.h>
-#include <unistd.h>
 #include <string.h>
 #include <stdlib.h>
 #include <sys/param.h>
@@ -35,39 +33,38 @@ int mount_init(void **context)
 	return 0;
 }
 
+int mount_reinit(void **context)
+{
+	return 0;
+}
+
 int mount_mount(struct autofs_point *ap, const char *root, const char *name, int name_len,
 		const char *what, const char *fstype, const char *options,
 		void *context)
 {
-	char *fullpath;
+	char fullpath[PATH_MAX];
 	char buf[MAX_ERR_BUF];
 	int err;
-	int rlen, status, existed = 1;
+	int len, status, existed = 1;
+
+	if (ap->flags & MOUNT_FLAG_REMOUNT)
+		return 0;
 
 	/* Root offset of multi-mount */
-	if (*name == '/' && name_len == 1) {
-		rlen = strlen(root);
-		name_len = 0;
-	/* Direct mount name is absolute path so don't use root */
-	} else if (*name == '/')
-		rlen = 0;
-	else
-		rlen = strlen(root);
-
-	fullpath = alloca(rlen + name_len + 2);
-	if (!fullpath) {
-		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-		error(ap->logopt, MODPREFIX "alloca: %s", estr);
-		return 1;
+	len = strlen(root);
+	if (root[len - 1] == '/') {
+		len = snprintf(fullpath, len, "%s", root);
+	} else if (*name == '/') {
+		/*
+		 * Direct or offset mount, name is absolute path so
+		 * don't use root (but with move mount changes root
+		 * is now the same as name).
+		 */
+		len = sprintf(fullpath, "%s", root);
+	} else {
+		len = sprintf(fullpath, "%s/%s", root, name);
 	}
-
-	if (name_len) {
-		if (rlen)
-			sprintf(fullpath, "%s/%s", root, name);
-		else
-			sprintf(fullpath, "%s", name);
-	} else
-		sprintf(fullpath, "%s", root);
+	fullpath[len] = '\0';
 
 	debug(ap->logopt, MODPREFIX "calling mkdir_path %s", fullpath);
 
@@ -82,39 +79,33 @@ int mount_mount(struct autofs_point *ap, const char *root, const char *name, int
 	if (!status)
 		existed = 0;
 
-	if (is_mounted(_PATH_MOUNTED, fullpath, MNTS_REAL)) {
-		error(ap->logopt,
-		      MODPREFIX "warning: %s is already mounted", fullpath);
-		return 0;
-	}
-
 	if (options && options[0]) {
 		debug(ap->logopt,
-		      MODPREFIX "calling mount -t %s " SLOPPY "-o %s %s %s",
+		      MODPREFIX "calling mount -t %s -o %s %s %s",
 		      fstype, options, what, fullpath);
 
-		err = spawn_mount(log_debug, "-t", fstype,
-			     SLOPPYOPT "-o", options, what, fullpath, NULL);
+		err = spawn_mount(ap->logopt, "-t", fstype,
+				  "-o", options, what, fullpath, NULL);
 	} else {
 		debug(ap->logopt, MODPREFIX "calling mount -t %s %s %s",
 		      fstype, what, fullpath);
-		err = spawn_mount(log_debug, "-t", fstype, what, fullpath, NULL);
+		err = spawn_mount(ap->logopt, "-t", fstype, what, fullpath, NULL);
 	}
 
 	if (err) {
-		msg(MODPREFIX "failed to mount %s (type %s) on %s",
+		info(ap->logopt, MODPREFIX "failed to mount %s (type %s) on %s",
 		     what, fstype, fullpath);
 
 		if (ap->type != LKP_INDIRECT)
 			return 1;
 
-		if ((!ap->ghost && name_len) || !existed)
+		if ((!(ap->flags & MOUNT_FLAG_GHOST) && name_len) || !existed)
 			rmdir_path(ap, fullpath, ap->dev);
 
 		return 1;
 	} else {
-		msg(MODPREFIX "mounted %s type %s on %s",
-		    what, fstype, fullpath);
+		debug(ap->logopt, MODPREFIX "mounted %s type %s on %s",
+		     what, fstype, fullpath);
 		return 0;
 	}
 }
