@@ -31,6 +31,7 @@
 
 #define MODULE_PARSE
 #include "automount.h"
+#include "nsswitch.h"
 
 #define MODPREFIX "parse(amd): "
 
@@ -129,6 +130,11 @@ int parse_init(int argc, const char *const *argv, void **context)
 	return 0;
 }
 
+int parse_reinit(int argc, const char *const *argv, void **context)
+{
+	return 0;
+}
+
 static struct substvar *add_lookup_vars(struct autofs_point *ap,
 					const char *key, int key_len,
 					struct map_source *source,
@@ -238,6 +244,11 @@ static int match_my_name(unsigned int logopt, const char *name, struct substvar
 			return 1;
 	}
 
+	if (!v || !v->val) {
+		error(logopt, "error: ${host} not set");
+		goto out;
+	}
+
 	/* Check if comparison value is an alias */
 
 	memset(&hints, 0, sizeof(hints));
@@ -274,7 +285,6 @@ static int match_my_name(unsigned int logopt, const char *name, struct substvar
 			error(logopt,
 			      "host address info lookup failed: %s\n",
 			      gai_strerror(ret));
-			freeaddrinfo(cni);
 			goto next;
 		}
 
@@ -285,7 +295,6 @@ static int match_my_name(unsigned int logopt, const char *name, struct substvar
 			error(logopt,
 			      "host address info lookup failed: %s\n",
 			      gai_strerror(ret));
-			freeaddrinfo(cni);
 			goto next;
 		}
 
@@ -901,9 +910,20 @@ static int do_auto_mount(struct autofs_point *ap, const char *name,
 {
 	char target[PATH_MAX + 1];
 
-	if (!entry->map_type)
+	if (!entry->map_type) {
+		if (strlen(entry->fs) > PATH_MAX) {
+			error(ap->logopt, MODPREFIX
+			     "error: fs option length is too long");
+			return 0;
+		}
 		strcpy(target, entry->fs);
-	else {
+	} else {
+		if (strlen(entry->fs) +
+		    strlen(entry->map_type) + 5 > PATH_MAX) {
+			error(ap->logopt, MODPREFIX
+			     "error: fs + maptype options length is too long");
+			return 0;
+		}
 		strcpy(target, entry->map_type);
 		strcat(target, ",amd:");
 		strcat(target, entry->fs);
@@ -920,10 +940,21 @@ static int do_link_mount(struct autofs_point *ap, const char *name,
 	const char *opts = (entry->opts && *entry->opts) ? entry->opts : NULL;
 	int ret;
 
-	if (entry->sublink)
+	if (entry->sublink) {
+		if (strlen(entry->sublink) > PATH_MAX) {
+			error(ap->logopt, MODPREFIX
+			     "error: sublink option length is too long");
+			return 0;
+		}
 		strcpy(target, entry->sublink);
-	else
+	} else {
+		if (strlen(entry->fs) > PATH_MAX) {
+			error(ap->logopt, MODPREFIX
+			     "error: fs option length is too long");
+			return 0;
+		}
 		strcpy(target, entry->fs);
+	}
 
 	if (!(flags & CONF_AUTOFS_USE_LOFS))
 		goto symlink;
@@ -1012,6 +1043,12 @@ static int do_nfs_mount(struct autofs_point *ap, const char *name,
 	unsigned int umount = 0;
 	int ret = 0;
 
+	if (strlen(entry->rhost) + strlen(entry->rfs) + 1 > PATH_MAX) {
+		error(ap->logopt, MODPREFIX
+		     "error: rhost + rfs options length is too long");
+		return 0;
+	}
+
 	strcpy(target, entry->rhost);
 	strcat(target, ":");
 	strcat(target, entry->rfs);
@@ -1096,6 +1133,7 @@ static int do_host_mount(struct autofs_point *ap, const char *name,
 	struct mapent *me;
 	const char *argv[2];
 	const char **pargv = NULL;
+	int status;
 	int argc = 0;
 	int ret = 1;
 
@@ -1137,8 +1175,8 @@ static int do_host_mount(struct autofs_point *ap, const char *name,
 	}
 
 	instance_mutex_lock();
-	lookup = open_lookup("hosts", MODPREFIX, NULL, argc, pargv);
-	if (!lookup) {
+	status = open_lookup("hosts", MODPREFIX, NULL, argc, pargv, &lookup);
+	if (status != NSS_STATUS_SUCCESS) {
 		debug(ap->logopt, "open lookup module hosts failed");
 		instance_mutex_unlock();
 		goto out;
@@ -1148,7 +1186,7 @@ static int do_host_mount(struct autofs_point *ap, const char *name,
 					 "hosts", "sun", argc, pargv);
 	if (!instance) {
 		instance = master_add_source_instance(source,
-				 "hosts", "sun", time(NULL), argc, pargv);
+				 "hosts", "sun", monotonic_time(NULL), argc, pargv);
 		if (!instance) {
 			error(ap->logopt, MODPREFIX
 			     "failed to create source instance for hosts map");
@@ -1193,13 +1231,12 @@ static unsigned int validate_auto_options(unsigned int logopt,
 	 * left blank the mount must be expected to fail so don't
 	 * report the error.
 	 */
-	if (!*entry->fs)
-		return 0;
-	else if (!entry->fs) {
+	if (!entry->fs) {
 		error(logopt, MODPREFIX
 		      "%s: file system not given", entry->type);
 		return 0;
-	}
+	} else if (!*entry->fs)
+		return 0;
 	return 1;
 }
 
@@ -1222,13 +1259,12 @@ static unsigned int validate_nfs_options(unsigned int logopt,
 	 * expected to fail.
 	 */
 	if (!entry->rfs || !*entry->rfs) {
-		if (!*entry->rfs)
+		if (entry->rfs && !*entry->rfs)
 			return 0;
 		/* Map option fs has been intentionally left blank */
 		if (entry->fs && !*entry->fs)
 			return 0;
-		if (entry->fs)
-			entry->rfs = strdup(entry->fs);
+		entry->rfs = strdup(entry->fs);
 		if (!entry->rfs) {
 			error(logopt, MODPREFIX
 			      "%s: remote file system not given", entry->type);
@@ -1252,24 +1288,22 @@ static unsigned int validate_generic_options(unsigned int logopt,
 	 * expected to fail so don't report the error.
 	 */
 	if (fstype != AMD_MOUNT_TYPE_LOFS) {
-		if (!*entry->dev)
-			return 0;
-		else if (!entry->dev) {
+		if (!entry->dev) {
 			error(logopt, MODPREFIX
 			      "%s: mount device not given", entry->type);
 			return 0;
-		}
-	} else {
-		if (!*entry->rfs)
+		} else if (!*entry->dev)
 			return 0;
-		else if (!entry->rfs) {
+	} else {
+		if (!entry->rfs) {
 			/*
 			 * Can't use entry->type as the mount type to reprot
 			 * the error since entry->type == "bind" not "lofs".
 			 */
 			error(logopt, "lofs: mount device not given");
 			return 0;
-		}
+		} else if (!*entry->rfs)
+			return 0;
 	}
 	if (entry->sublink && !entry->fs) {
 		error(logopt, MODPREFIX
@@ -1304,13 +1338,12 @@ static unsigned int validate_host_options(unsigned int logopt,
 	 * if it isn't given in the map entry. Don't report an error
 	 * if it has been left empty since it's expected to fail.
 	 */
-	if (!*entry->rhost)
-		return 0;
-	else if (!entry->rhost) {
+	if (!entry->rhost) {
 		error(logopt, MODPREFIX
 		      "%s: remote host name not given", entry->type);
 		return 0;
-	}
+	} else if (!*entry->rhost)
+		return 0;
 	return 1;
 }
 
@@ -1381,7 +1414,7 @@ static int amd_mount(struct autofs_point *ap, const char *name,
 
 	default:
 		info(ap->logopt,
-		     MODPREFIX "unkown file system type %x", fstype);
+		     MODPREFIX "unknown file system type %x", fstype);
 		break;
 	}
 
@@ -1603,7 +1636,7 @@ static struct amd_entry *dup_defaults_entry(struct amd_entry *defaults)
 			entry->rfs = tmp;
 	}
 
-	if (defaults->rhost && *defaults->rfs) {
+	if (defaults->rhost && *defaults->rhost) {
 		tmp = strdup(defaults->rhost);
 		if (tmp)
 			entry->rhost = tmp;
@@ -1769,8 +1802,13 @@ static struct amd_entry *get_defaults_entry(struct autofs_point *ap,
 		char *expand;
 		if (!expand_selectors(ap, defaults, &expand, sv))
 			goto out;
-		if (amd_parse_list(ap, expand, &dflts, &sv))
+		if (amd_parse_list(ap, expand, &dflts, &sv)) {
+			error(ap->logopt, MODPREFIX
+			     "failed to parse defaults entry, "
+			     "attempting to use internal default");
+			free(expand);
 			goto out;
+		}
 		entry = select_default_entry(ap, &dflts, sv);
 		if (!entry->map_type) {
 			/*
