@@ -232,27 +232,15 @@ int expandsunent(const char *src, char *dst, const char *key,
 	return len;
 }
 
-int parse_init(int argc, const char *const *argv, void **context)
+static int do_init(int argc, const char *const *argv, struct parse_context *ctxt)
 {
-	struct parse_context *ctxt;
-	char buf[MAX_ERR_BUF];
 	char *noptstr, *def, *val, *macros, *gbl_options;
-	const char *xopt;
+	char buf[MAX_ERR_BUF];
 	int optlen, len, offset;
+	const char *xopt;
 	int i, bval;
 	unsigned int append_options;
 
-	/* Set up context and escape chain */
-
-	if (!(ctxt = (struct parse_context *) malloc(sizeof(struct parse_context)))) {
-		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-		logerr(MODPREFIX "malloc: %s", estr);
-		*context = NULL;
-		return 1;
-	}
-	*context = (void *) ctxt;
-
-	*ctxt = default_context;
 	optlen = 0;
 
 	/* Look for options and capture, and create new defines if we need to */
@@ -357,9 +345,7 @@ int parse_init(int argc, const char *const *argv, void **context)
 			}
 			if (!noptstr) {
 				char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-				kill_context(ctxt);
 				logerr(MODPREFIX "%s", estr);
-				*context = NULL;
 				return 1;
 			}
 			ctxt->optstr = noptstr;
@@ -391,9 +377,36 @@ int parse_init(int argc, const char *const *argv, void **context)
 		}
 	}
 options_done:
+
 	debug(LOGOPT_NONE,
 	      MODPREFIX "init gathered global options: %s", ctxt->optstr);
 
+	return 0;
+}
+
+int parse_init(int argc, const char *const *argv, void **context)
+{
+	struct parse_context *ctxt;
+	char buf[MAX_ERR_BUF];
+
+	*context = NULL;
+
+	/* Set up context and escape chain */
+
+	ctxt = (struct parse_context *) malloc(sizeof(struct parse_context));
+	if (!ctxt) {
+		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
+		logerr(MODPREFIX "malloc: %s", estr);
+		return 1;
+	}
+
+	*ctxt = default_context;
+
+	if (do_init(argc, argv, ctxt)) {
+		free(ctxt);
+		return 1;
+	}
+
 	/* We only need this once.  NFS mounts are so common that we cache
 	   this module. */
 	instance_mutex_lock();
@@ -404,12 +417,39 @@ options_done:
 			init_ctr++;
 		} else {
 			kill_context(ctxt);
-			*context = NULL;
 			instance_mutex_unlock();
 			return 1;
 		}
 	}
 	instance_mutex_unlock();
+
+	*context = (void *) ctxt;
+
+	return 0;
+}
+
+int parse_reinit(int argc, const char *const *argv, void **context)
+{
+	struct parse_context *ctxt = (struct parse_context *) *context;
+	struct parse_context *new;
+	char buf[MAX_ERR_BUF];
+
+	new = (struct parse_context *) malloc(sizeof(struct parse_context));
+	if (!new) {
+		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
+		logerr(MODPREFIX "malloc: %s", estr);
+		return 1;
+	}
+
+	*new = default_context;
+
+	if (do_init(argc, argv, new))
+		return 1;
+
+	kill_context(ctxt);
+
+	*context = (void *) new;
+
 	return 0;
 }
 
@@ -509,7 +549,8 @@ static int sun_mount(struct autofs_point *ap, const char *root,
 				fstype[typelen] = '\0';
 			} else if (_strncmp("nonstrict", cp, 9) == 0) {
 				nonstrict = 1;
-			} else if (_strncmp("strict", cp, 6) == 0) {
+			} else if (_strncmp("strict", cp, 6) == 0 &&
+				   comma - cp == 6) {
 				nonstrict = 0;
 			} else if (_strncmp("nobrowse", cp, 8) == 0 ||
 				   _strncmp("browse", cp, 6) == 0 ||
@@ -756,6 +797,8 @@ update_offset_entry(struct autofs_point *ap, const char *name,
 
 	mc = source->mc;
 
+	memset(m_mapent, 0, MAPENT_MAX_LEN + 1);
+
 	/* Internal hosts map may have loc == NULL */
 	if (!*path) {
 		error(ap->logopt,
@@ -782,7 +825,7 @@ update_offset_entry(struct autofs_point *ap, const char *name,
 	if (*myoptions)
 		m_options_len = strlen(myoptions) + 2;
 
-	m_mapent_len = strlen(loc);
+	m_mapent_len = loc ? strlen(loc) : 0;
 	if (m_mapent_len + m_options_len > MAPENT_MAX_LEN) {
 		error(ap->logopt, MODPREFIX "multi mount mapent too long");
 		return CHE_FAIL;
@@ -793,16 +836,20 @@ update_offset_entry(struct autofs_point *ap, const char *name,
 		strcat(m_mapent, myoptions);
 		if (loc) {
 			strcat(m_mapent, " ");
-			strcat(m_mapent, loc);
+			if (loc)
+				strcat(m_mapent, loc);
 		}
-	} else
-		strcpy(m_mapent, loc);
+	} else {
+		if (loc)
+			strcpy(m_mapent, loc);
+	}
 
 	ret = cache_update_offset(mc, name, m_key, m_mapent, age);
-	if (ret == CHE_DUPLICATE)
+	if (ret == CHE_DUPLICATE) {
 		warn(ap->logopt, MODPREFIX
 		     "syntax error or duplicate offset %s -> %s", path, loc);
-	else if (ret == CHE_FAIL)
+		ret = CHE_OK;
+	} else if (ret == CHE_FAIL)
 		debug(ap->logopt, MODPREFIX
 		      "failed to update multi-mount offset %s -> %s", path, m_mapent);
 	else {
@@ -923,9 +970,15 @@ static int parse_mapent(const char *ent, char *g_options, char **options, char *
 	l = chunklen(p, check_colon(p));
 	loc = dequote(p, l, logopt);
 	if (!loc) {
-		warn(logopt, MODPREFIX "possible missing location");
-		free(myoptions);
-		return 0;
+		if (strstr(myoptions, "fstype=autofs") &&
+		    strstr(myoptions, "hosts")) {
+			warn(logopt, MODPREFIX "possible missing location");
+			free(myoptions);
+			return 0;
+		}
+		*options = myoptions;
+		*location = NULL;
+		return (p - ent);
 	}
 
 	/* Location can't begin with a '/' */
@@ -953,10 +1006,15 @@ static int parse_mapent(const char *ent, char *g_options, char **options, char *
 		l = chunklen(p, check_colon(p));
 		ent_chunk = dequote(p, l, logopt);
 		if (!ent_chunk) {
-			warn(logopt, MODPREFIX "null location or out of memory");
-			free(myoptions);
-			free(loc);
-			return 0;
+			if (strstr(myoptions, "fstype=autofs") &&
+			    strstr(myoptions, "hosts")) {
+				warn(logopt, MODPREFIX
+				     "null location or out of memory");
+				free(myoptions);
+				free(loc);
+				return 0;
+			}
+			goto next;
 		}
 
 		/* Location can't begin with a '/' */
@@ -992,7 +1050,7 @@ static int parse_mapent(const char *ent, char *g_options, char **options, char *
 		strcat(loc, ent_chunk);
 
 		free(ent_chunk);
-
+next:
 		p += l;
 		p = skipspace(p);
 	}
@@ -1093,7 +1151,9 @@ static int mount_subtree(struct autofs_point *ap, struct mapent *me,
 				cache_delete_offset_list(me->mc, name);
 				return 1;
 			}
-			ro_len = strlen(ro_loc);
+			ro_len = 0;
+			if (ro_loc)
+				ro_len = strlen(ro_loc);
 
 			tmp = alloca(mnt_root_len + 2);
 			strcpy(tmp, mnt_root);
@@ -1104,7 +1164,8 @@ static int mount_subtree(struct autofs_point *ap, struct mapent *me,
 			rv = sun_mount(ap, root, name, namelen, ro_loc, ro_len, myoptions, ctxt);
 
 			free(myoptions);
-			free(ro_loc);
+			if (ro_loc)
+				free(ro_loc);
 		}
 
 		if (ro && rv == 0) {
@@ -1214,12 +1275,12 @@ int parse_mount(struct autofs_point *ap, const char *name,
 	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cur_state);
 	macro_lock();
 
-	ctxt->subst = addstdenv(ctxt->subst);
+	ctxt->subst = addstdenv(ctxt->subst, NULL);
 
 	mapent_len = expandsunent(mapent, NULL, name, ctxt->subst, slashify);
 	if (mapent_len == 0) {
 		error(ap->logopt, MODPREFIX "failed to expand map entry");
-		ctxt->subst = removestdenv(ctxt->subst);
+		ctxt->subst = removestdenv(ctxt->subst, NULL);
 		macro_unlock();
 		pthread_setcancelstate(cur_state, NULL);
 		return 1;
@@ -1229,7 +1290,7 @@ int parse_mount(struct autofs_point *ap, const char *name,
 	if (!pmapent) {	
 		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
 		logerr(MODPREFIX "alloca: %s", estr);
-		ctxt->subst = removestdenv(ctxt->subst);
+		ctxt->subst = removestdenv(ctxt->subst, NULL);
 		macro_unlock();
 		pthread_setcancelstate(cur_state, NULL);
 		return 1;
@@ -1237,7 +1298,7 @@ int parse_mount(struct autofs_point *ap, const char *name,
 	pmapent[mapent_len] = '\0';
 
 	expandsunent(mapent, pmapent, name, ctxt->subst, slashify);
-	ctxt->subst = removestdenv(ctxt->subst);
+	ctxt->subst = removestdenv(ctxt->subst, NULL);
 
 	macro_unlock();
 	pthread_setcancelstate(cur_state, NULL);
@@ -1343,26 +1404,51 @@ int parse_mount(struct autofs_point *ap, const char *name,
 			strcat(m_root, name);
 		}
 
+		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cur_state);
 		cache_readlock(mc);
 		me = cache_lookup_distinct(mc, name);
-		if (me) {
-			/* So we know we're the multi-mount root */
-			if (!me->multi)
-				me->multi = me;
-		}
-
 		if (!me) {
 			free(options);
 			cache_unlock(mc);
+			pthread_setcancelstate(cur_state, NULL);
 			error(ap->logopt,
 			      MODPREFIX "can't find multi root %s", name);
 			return 1;
 		}
 
+		cache_multi_writelock(me);
+		/* So we know we're the multi-mount root */
+		if (!me->multi)
+			me->multi = me;
+		else {
+			/*
+			 * The amd host mount type assumes the lookup name
+			 * is the host name for the host mount but amd uses
+			 * ${rhost} for this.
+			 *
+			 * This introduces the possibility of multiple
+			 * concurrent mount requests since constructing a
+			 * mount tree that isn't under the lookup name can't
+			 * take advantage of the kernel queuing of other
+			 * concurrent lookups while the mount tree is
+			 * constructed.
+			 *
+			 * Consequently multi-mount updates (currently only
+			 * done for the internal hosts map which the amd
+			 * parser also uses for its hosts map) can't be
+			 * allowed for amd mounts.
+			 */
+			if (source->flags & MAP_FLAG_FORMAT_AMD) {
+				free(options);
+				cache_multi_unlock(me);
+				cache_unlock(mc);
+				pthread_setcancelstate(cur_state, NULL);
+				return 0;
+			}
+		}
+
 		age = me->age;
 
-		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cur_state);
-		cache_multi_writelock(me);
 		/* It's a multi-mount; deal with it */
 		do {
 			char *path, *myoptions, *loc;
@@ -1395,16 +1481,13 @@ int parse_mount(struct autofs_point *ap, const char *name,
 
 			l = parse_mapent(p, options, &myoptions, &loc, ap->logopt);
 			if (!l) {
-				if (!(strstr(myoptions, "fstype=autofs") &&
-				      strstr(myoptions, "hosts"))) {
-					cache_delete_offset_list(mc, name);
-					cache_multi_unlock(me);
-					cache_unlock(mc);
-					free(path);
-					free(options);
-					pthread_setcancelstate(cur_state, NULL);
-					return 1;
-				}
+				cache_delete_offset_list(mc, name);
+				cache_multi_unlock(me);
+				cache_unlock(mc);
+				free(path);
+				free(options);
+				pthread_setcancelstate(cur_state, NULL);
+				return 1;
 			}
 
 			p += l;
@@ -1425,12 +1508,14 @@ int parse_mount(struct autofs_point *ap, const char *name,
 				free(path);
 				free(options);
 				free(myoptions);
-				free(loc);
+				if (loc)
+					free(loc);
 				pthread_setcancelstate(cur_state, NULL);
 				return 1;
 			}
 
-			free(loc);
+			if (loc)
+				free(loc);
 			free(path);
 			free(myoptions);
 		} while (*p == '/' || (*p == '"' && *(p + 1) == '/'));
@@ -1561,7 +1646,10 @@ int parse_mount(struct autofs_point *ap, const char *name,
 		 */
 		if ((strstr(options, "fstype=autofs") &&
 		     strstr(options, "hosts"))) {
-			loc = NULL;
+			if (loc) {
+				free(loc);
+				loc = NULL;
+			}
 			loclen = 0;
 		} else {
 			loclen = strlen(loc);
@@ -1585,7 +1673,8 @@ int parse_mount(struct autofs_point *ap, const char *name,
 			rv = sun_mount(ap, ap->path, name, name_len,
 				       loc, loclen, options, ctxt);
 
-		free(loc);
+		if (loc)
+			free(loc);
 		free(options);
 		pthread_setcancelstate(cur_state, NULL);
 	}
